"""Tests for deprecated methods in ``glotaran.model.model``."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from glotaran.deprecation.deprecation_utils import GlotaranDeprectedApiError
from glotaran.testing.model_generators import SimpleModelGenerator

if TYPE_CHECKING:
    from glotaran.model import Model


@pytest.fixture(scope="module")
def dummy_model() -> Model:
    """Minimal model instance for testing."""
    generator = SimpleModelGenerator(
        rates=[300e-3],
        k_matrix="parallel",
    )
    return generator.model


def test_model_model_dimension(dummy_model: Model):
    """Raise ``GlotaranApiDeprecationWarning``."""
    expected = (
        "Usage of 'Model.model_dimension' was deprecated, "
        "use \"Scheme.model_dimensions['<dataset_name>']\" instead.\n"
        "It wasn't possible to restore the original behavior of this usage "
        "(mostlikely due to an object hierarchy change)."
        "This usage change message won't be show as of version: '0.7.0'."
    )

    with pytest.raises(GlotaranDeprectedApiError) as excinfo:
        dummy_model.model_dimension

    assert str(excinfo.value) == expected


def test_model_global_dimension(dummy_model: Model):
    """Raise ``GlotaranApiDeprecationWarning``."""
    expected = (
        "Usage of 'Model.global_dimension' was deprecated, "
        "use \"Scheme.global_dimensions['<dataset_name>']\" instead.\n"
        "It wasn't possible to restore the original behavior of this usage "
        "(mostlikely due to an object hierarchy change)."
        "This usage change message won't be show as of version: '0.7.0'."
    )

    with pytest.raises(GlotaranDeprectedApiError) as excinfo:
        dummy_model.global_dimension

    assert str(excinfo.value) == expected



"""This package contains the decay megacomplex item."""
from __future__ import annotations

from typing import List

import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.initial_concentration import InitialConcentration
from glotaran.builtin.megacomplexes.decay.irf import Irf
from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.builtin.megacomplexes.decay.k_matrix import KMatrix
from glotaran.builtin.megacomplexes.decay.util import decay_matrix_implementation
from glotaran.builtin.megacomplexes.decay.util import retrieve_decay_associated_data
from glotaran.builtin.megacomplexes.decay.util import retrieve_irf
from glotaran.builtin.megacomplexes.decay.util import retrieve_species_associated_data
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import megacomplex


@megacomplex(
    dimension="time",
    model_items={
        "k_matrix": List[KMatrix],
    },
    properties={},
    dataset_model_items={
        "initial_concentration": {"type": InitialConcentration, "allow_none": True},
        "irf": {"type": Irf, "allow_none": True},
    },
    register_as="decay",
)
class DecayMegacomplex(Megacomplex):
    """A Megacomplex with one or more K-Matrices."""

    def has_k_matrix(self) -> bool:
        return len(self.k_matrix) != 0

    def full_k_matrix(self, model=None):
        full_k_matrix = None
        for k_matrix in self.k_matrix:
            if model:
                k_matrix = model.k_matrix[k_matrix]
            if full_k_matrix is None:
                full_k_matrix = k_matrix
            # If multiple k matrices are present, we combine them
            else:
                full_k_matrix = full_k_matrix.combine(k_matrix)
        return full_k_matrix

    @property
    def involved_compartments(self):
        return self.full_k_matrix().involved_compartments() if self.full_k_matrix() else []

    def index_dependent(self, dataset_model: DatasetModel) -> bool:
        return (
            isinstance(dataset_model.irf, IrfMultiGaussian)
            and dataset_model.irf.is_index_dependent()
        )

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        indices: dict[str, int],
        **kwargs,
    ):
        if dataset_model.initial_concentration is None:
            raise ModelError(
                f'No initial concentration specified in dataset "{dataset_model.label}"'
            )
        initial_concentration = dataset_model.initial_concentration.normalized()

        k_matrix = self.full_k_matrix()

        # we might have more species in the model then in the k matrix
        species = [
            comp
            for comp in initial_concentration.compartments
            if comp in k_matrix.involved_compartments()
        ]

        # the rates are the eigenvalues of the k matrix
        rates = k_matrix.rates(initial_concentration)

        global_dimension = dataset_model.get_global_dimension()
        global_index = indices.get(global_dimension)
        global_axis = dataset_model.get_global_axis()
        model_axis = dataset_model.get_model_axis()

        # init the matrix
        size = (model_axis.size, rates.size)
        matrix = np.zeros(size, dtype=np.float64)

        decay_matrix_implementation(
            matrix, rates, global_index, global_axis, model_axis, dataset_model
        )

        if not np.all(np.isfinite(matrix)):
            raise ValueError(
                f"Non-finite concentrations for K-Matrix '{k_matrix.label}':\n"
                f"{k_matrix.matrix_as_markdown(fill_parameters=True)}"
            )

        # apply A matrix
        matrix = matrix @ k_matrix.a_matrix(initial_concentration)

        # done
        return species, matrix

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        global_dimension = dataset_model.get_global_dimension()
        name = "images" if global_dimension == "pixel" else "spectra"

        species_dimension = "decay_species" if as_global else "species"
        if species_dimension not in dataset.coords:
            # We are the first Decay complex called and add SAD for all decay megacomplexes
            retrieve_species_associated_data(
                dataset_model,
                dataset,
                species_dimension,
                global_dimension,
                name,
                is_full_model,
                as_global,
            )
        if isinstance(dataset_model.irf, IrfMultiGaussian) and "irf" not in dataset:
            retrieve_irf(dataset_model, dataset, global_dimension)

        if not is_full_model:
            multiple_complexes = (
                len([m for m in dataset_model.megacomplex if isinstance(m, DecayMegacomplex)]) > 1
            )
            retrieve_decay_associated_data(
                self,
                dataset_model,
                dataset,
                global_dimension,
                name,
                multiple_complexes,
            )



"""The parameter class."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import asteval
import numpy as np
from numpy.typing._array_like import _SupportsArray

from glotaran.utils.sanitize import sanitize_parameter_list

if TYPE_CHECKING:
    from typing import Any

    from glotaran.parameter import ParameterGroup

RESERVED_LABELS: list[str] = list(asteval.make_symbol_table().keys()) + ["group"]


class Keys:
    """Keys for parameter options."""

    EXPR = "expr"
    MAX = "max"
    MIN = "min"
    NON_NEG = "non-negative"
    VARY = "vary"


PARAMETER_EXPRESION_REGEX = re.compile(r"\$(?P<parameter_expression>[\w\d\.]+)((?![\w\d\.]+)|$)")
"""A regexpression to find and replace parameter names in expressions."""
VALID_LABEL_REGEX = re.compile(r"\W", flags=re.ASCII)
"""A regexpression to validate labels."""


class Parameter(_SupportsArray):
    """A parameter for optimization."""

    def __init__(
        self,
        label: str = None,
        full_label: str = None,
        expression: str = None,
        maximum: int | float = np.inf,
        minimum: int | float = -np.inf,
        non_negative: bool = False,
        value: float | int = np.nan,
        vary: bool = True,
    ):
        """Optimization Parameter supporting numpy array operations.

        Parameters
        ----------
        label : str
            The label of the parameter., by default None
        full_label : str
            The label of the parameter with its path in a parameter group prepended.
            , by default None
        expression : str
            Expression to calculate the parameters value from,
            e.g. if used in relation to another parameter. , by default None
        maximum : int
            Upper boundary for the parameter to be varied to., by default np.inf
        minimum : int
            Lower boundary for the parameter to be varied to., by default -np.inf
        non_negative : bool
            Whether the parameter should always be bigger than zero., by default False
        value : float
            Value of the parameter, by default np.nan
        vary : bool
            Whether the parameter should be changed during optimization or not.
            , by default True
        """
        self.label = label
        self.full_label = full_label or ""
        self.expression = expression
        self.maximum = maximum
        self.minimum = minimum
        self.non_negative = non_negative
        self.standard_error = 0.0
        self.value = value
        self.vary = vary

        self._transformed_expression: str | None = None

    @staticmethod
    def valid_label(label: str) -> bool:
        """Check if a label is a valid label for :class:`Parameter`.

        Parameters
        ----------
        label : str
            The label to validate.

        Returns
        -------
        bool
            Whether the label is valid.

        """
        return VALID_LABEL_REGEX.search(label) is None and label not in RESERVED_LABELS

    @classmethod
    def from_list_or_value(
        cls,
        value: int | float | list,
        default_options: dict = None,
        label: str = None,
    ) -> Parameter:
        """Create a parameter from a list or numeric value.

        Parameters
        ----------
        value : int | float | list
            The list or numeric value.
        default_options : dict
            A dictionary of default options.
        label : str
            The label of the parameter.

        Returns
        -------
        Parameter
            The created :class:`Parameter`.
        """
        param = cls(label=label)
        options = None

        if not isinstance(value, list):
            param.value = value

        else:
            values = sanitize_parameter_list(value)
            param.label = _retrieve_item_from_list_by_type(values, str, label)
            param.value = float(_retrieve_item_from_list_by_type(values, (int, float), 0))
            options = _retrieve_item_from_list_by_type(values, dict, None)

        if default_options:
            param._set_options_from_dict(default_options)

        if options:
            param._set_options_from_dict(options)
        return param

    def set_from_group(self, group: ParameterGroup):
        """Set all values of the parameter to the values of the corresponding parameter in the group.

        Notes
        -----
        For internal use.

        Parameters
        ----------
        group : ParameterGroup
            The :class:`glotaran.parameter.ParameterGroup`.
        """
        p = group.get(self.full_label)
        self.expression = p.expression
        self.maximum = p.maximum
        self.minimum = p.minimum
        self.non_negative = p.non_negative
        self.standard_error = p.standard_error
        self.value = p.value
        self.vary = p.vary

    def _set_options_from_dict(self, options: dict):
        """Set the parameter's options from a dictionary.

        Parameters
        ----------
        options : dict
            A dictionary containing parameter options.
        """
        if Keys.EXPR in options:
            self.expression = options[Keys.EXPR]
        if Keys.NON_NEG in options:
            self.non_negative = options[Keys.NON_NEG]
        if Keys.MAX in options:
            self.maximum = options[Keys.MAX]
        if Keys.MIN in options:
            self.minimum = options[Keys.MIN]
        if Keys.VARY in options:
            self.vary = options[Keys.VARY]

    @property
    def label(self) -> str | None:
        """Label of the parameter.

        Returns
        -------
        str
            The label.
        """
        return self._label

    @label.setter
    def label(self, label: str | None):
        # ensure that label is str | None even if an int is passed
        label = None if label is None else str(label)
        if label is not None and not Parameter.valid_label(label):
            raise ValueError(f"'{label}' is not a valid group label.")
        self._label = label

    @property
    def full_label(self) -> str:
        """Label of the parameter with its path in a parameter group prepended.

        Returns
        -------
        str
            The full label.
        """
        return self._full_label

    @full_label.setter
    def full_label(self, full_label: str):
        self._full_label = str(full_label)

    @property
    def non_negative(self) -> bool:
        r"""Indicate if the parameter is non-negativ.

        If true, the parameter will be transformed with :math:`p' = \log{p}` and
        :math:`p = \exp{p'}`.

        Notes
        -----
        Always `False` if `expression` is not `None`.

        Returns
        -------
        bool
            Whether the parameter is non-negativ.
        """
        return self._non_negative if self.expression is None else False

    @non_negative.setter
    def non_negative(self, non_negative: bool):
        self._non_negative = non_negative

    @property
    def vary(self) -> bool:
        """Indicate if the parameter should be optimized.

        Notes
        -----
        Always `False` if `expression` is not `None`.

        Returns
        -------
        bool
            Whether the parameter should be optimized.
        """
        return self._vary if self.expression is None else False

    @vary.setter
    def vary(self, vary: bool):
        self._vary = vary

    @property
    def maximum(self) -> float:
        """Upper bound of the parameter.

        Returns
        -------
        float
            The upper bound of the parameter.
        """
        return self._maximum

    @maximum.setter
    def maximum(self, maximum: int | float):
        if not isinstance(maximum, float):
            try:
                maximum = float(maximum)
            except Exception:
                raise TypeError(
                    "Parameter maximum must be numeric."
                    + f"'{self.full_label}' has maximum '{maximum}' of type '{type(maximum)}'"
                )

        self._maximum = maximum

    @property
    def minimum(self) -> float:
        """Lower bound of the parameter.

        Returns
        -------
        float

            The lower bound of the parameter.
        """
        return self._minimum

    @minimum.setter
    def minimum(self, minimum: int | float):
        if not isinstance(minimum, float):
            try:
                minimum = float(minimum)
            except Exception:
                raise TypeError(
                    "Parameter minimum must be numeric."
                    + f"'{self.full_label}' has minimum '{minimum}' of type '{type(minimum)}'"
                )

        self._minimum = minimum

    @property
    def expression(self) -> str | None:
        """Expression to calculate the parameters value from.

        This can used to set a relation to another parameter.

        Returns
        -------
        str | None
            The expression.
        """
        return self._expression

    @expression.setter
    def expression(self, expression: str | None):
        self._expression = expression
        self._transformed_expression = None

    @property
    def transformed_expression(self) -> str | None:
        """Expression of the parameter transformed for evaluation within a `ParameterGroup`.

        Returns
        -------
        str | None
            The transformed expression.
        """
        if self.expression is not None and self._transformed_expression is None:
            self._transformed_expression = PARAMETER_EXPRESION_REGEX.sub(
                r"group.get('\g<parameter_expression>').value", self.expression
            )
        return self._transformed_expression

    @property
    def standard_error(self) -> float:
        """Standard error of the optimized parameter.

        Returns
        -------
        float
            The standard error of the parameter.
        """  # noqa: D401
        return self._stderr

    @standard_error.setter
    def standard_error(self, standard_error: float):
        self._stderr = standard_error

    @property
    def value(self) -> float:
        """Value of the parameter.

        Returns
        -------
        float
            The value of the parameter.
        """
        return self._value

    @value.setter
    def value(self, value: int | float):
        if not isinstance(value, float) and value is not np.nan:
            try:
                value = float(value)
            except Exception:
                raise TypeError(
                    "Parameter value must be numeric."
                    + f"'{self.full_label}' has value '{value}' of type '{type(value)}'"
                )

        self._value = value

    def get_value_and_bounds_for_optimization(self) -> tuple[float, float, float]:
        """Get the parameter value and bounds with expression and non-negative constraints applied.

        Returns
        -------
        tuple[float, float, float]
            A tuple containing the value, the lower and the upper bound.
        """
        value = self.value
        minimum = self.minimum
        maximum = self.maximum

        if self.non_negative:
            value = _log_value(value)
            minimum = _log_value(minimum)
            maximum = _log_value(maximum)

        return value, minimum, maximum

    def set_value_from_optimization(self, value: float):
        """Set the value from an optimization result and reverses non-negative transformation.

        Parameters
        ----------
        value : float
            Value from optimization.
        """
        self.value = np.exp(value) if self.non_negative else value

    def __getstate__(self):
        """Get state for pickle."""
        return (
            self.label,
            self.full_label,
            self.expression,
            self.maximum,
            self.minimum,
            self.non_negative,
            self.standard_error,
            self.value,
            self.vary,
        )

    def __setstate__(self, state):
        """Set state from pickle."""
        (
            self.label,
            self.full_label,
            self.expression,
            self.maximum,
            self.minimum,
            self.non_negative,
            self.standard_error,
            self.value,
            self.vary,
        ) = state

    def __repr__(self):
        """Representation used by repl and tracebacks."""
        return (
            f"{type(self).__name__}(label={self.label!r}, value={self.value!r},"
            f" expression={self.expression!r}, vary={self.vary!r})"
        )

    def __array__(self):
        """array"""  # noqa: D400, D403
        return np.array(float(self._value), dtype=float)

    def __str__(self) -> str:
        """Representation used by print and str."""
        return (
            f"__{self.label}__: _Value_: {self.value}, _StdErr_: {self.standard_error}, _Min_:"
            f" {self.minimum}, _Max_: {self.maximum}, _Vary_: {self.vary},"
            f" _Non-Negative_: {self.non_negative}, _Expr_: {self.expression}"
        )

    def __abs__(self):
        """abs"""  # noqa: D400, D403
        return abs(self._value)

    def __neg__(self):
        """neg"""  # noqa: D400, D403
        return -self._value

    def __pos__(self):
        """positive"""  # noqa: D400, D403
        return +self._value

    def __int__(self):
        """int"""  # noqa: D400, D403
        return int(self._value)

    def __float__(self):
        """float"""  # noqa: D400, D403
        return float(self._value)

    def __trunc__(self):
        """trunc"""  # noqa: D400, D403
        return self._value.__trunc__()

    def __add__(self, other):
        """+"""  # noqa: D400
        return self._value + other

    def __sub__(self, other):
        """-"""  # noqa: D400
        return self._value - other

    def __truediv__(self, other):
        """/"""  # noqa: D400
        return self._value / other

    def __floordiv__(self, other):
        """//"""  # noqa: D400
        return self._value // other

    def __divmod__(self, other):
        """divmod"""  # noqa: D400, D403
        return divmod(self._value, other)

    def __mod__(self, other):
        """%"""  # noqa: D400
        return self._value % other

    def __mul__(self, other):
        """*"""  # noqa: D400
        return self._value * other

    def __pow__(self, other):
        """**"""  # noqa: D400
        return self._value ** other

    def __gt__(self, other):
        """>"""  # noqa: D400
        return self._value > other

    def __ge__(self, other):
        """>="""  # noqa: D400
        return self._value >= other

    def __le__(self, other):
        """<="""  # noqa: D400
        return self._value <= other

    def __lt__(self, other):
        """<"""  # noqa: D400
        return self._value < other

    def __eq__(self, other):
        """=="""  # noqa: D400
        return self._value == other

    def __ne__(self, other):
        """!="""  # noqa: D400
        return self._value != other

    def __radd__(self, other):
        """+ (right)"""  # noqa: D400
        return other + self._value

    def __rtruediv__(self, other):
        """/ (right)"""  # noqa: D400
        return other / self._value

    def __rdivmod__(self, other):
        """divmod (right)"""  # noqa: D400, D403
        return divmod(other, self._value)

    def __rfloordiv__(self, other):
        """// (right)"""  # noqa: D400
        return other // self._value

    def __rmod__(self, other):
        """% (right)"""  # noqa: D400
        return other % self._value

    def __rmul__(self, other):
        """* (right)"""  # noqa: D400
        return other * self._value

    def __rpow__(self, other):
        """** (right)"""  # noqa: D400
        return other ** self._value

    def __rsub__(self, other):
        """- (right)"""  # noqa: D400
        return other - self._value


def _log_value(value: float) -> float:
    """Get the logarithm of a value.

    Performs a check for edge cases and migitates numerical issues.

    Parameters
    ----------
    value : float
        The initial value.

    Returns
    -------
    float
        The logarithm of the value.
    """
    if not np.isfinite(value):
        return value
    if value == 1:
        value += 1e-10
    return np.log(value)


def _retrieve_item_from_list_by_type(
    item_list: list, item_type: type | tuple[type, ...], default: Any
) -> Any:
    """Retrieve an item from list which matches a given type.

    Parameters
    ----------
    item_list : list
        The list to retrieve from.
    item_type : type | tuple[type, ...]
        The item type or tuple of types to match.
    default : Any
        Returned if no item matches.

    Returns
    -------
    Any

    """
    tmp = list(filter(lambda x: isinstance(x, item_type), item_list))
    if not tmp:
        return default
    item_list.remove(tmp[0])
    return tmp[0]



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

import glotaran.builtin.io.yml.yml as yml_module
from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.io import load_model
from glotaran.io import load_scheme

if TYPE_CHECKING:
    from typing import Any

    from _pytest.monkeypatch import MonkeyPatch


@pytest.mark.parametrize(
    "model_yml_str, expected_nr_of_warnings, expected_key, expected_value",
    (
        ("type: kinetic-spectrum", 1, "default_megacomplex", "decay"),
        ("type: spectral-model", 1, "default_megacomplex", "spectral"),
        ("default-megacomplex: decay", 1, "default_megacomplex", "decay"),
        ("default-megacomplex: spectral", 1, "default_megacomplex", "spectral"),
        (
            dedent(
                """
                spectral_relations:
                    - compartment: s1
                    - compartment: s2
                """
            ),
            3,
            "clp_relations",
            [{"source": "s1"}, {"source": "s2"}],
        ),
        (
            dedent(
                """
                relations:
                    - compartment: s1
                    - compartment: s2
                """
            ),
            3,
            "clp_relations",
            [{"source": "s1"}, {"source": "s2"}],
        ),
        (
            dedent(
                """
                spectral_constraints:
                    - compartment: s1
                    - compartment: s2
                """
            ),
            3,
            "clp_constraints",
            [{"target": "s1"}, {"target": "s2"}],
        ),
        (
            dedent(
                """
                constraints:
                    - compartment: s1
                    - compartment: s2
                """
            ),
            3,
            "clp_constraints",
            [{"target": "s1"}, {"target": "s2"}],
        ),
        (
            dedent(
                """
                equal_area_penalties:
                    - type: equal_area
                """
            ),
            1,
            "clp_area_penalties",
            [{"type": "equal_area"}],
        ),
        (
            dedent(
                """
                irf:
                    irf1:
                        center_dispersion: [cdc1]

                """
            ),
            1,
            "irf",
            {"irf1": {"center_dispersion_coefficients": ["cdc1"]}},
        ),
        (
            dedent(
                """
                irf:
                    irf1:
                        "width_dispersion": [wdc1]

                """
            ),
            1,
            "irf",
            {"irf1": {"width_dispersion_coefficients": ["wdc1"]}},
        ),
    ),
    ids=(
        "type: kinetic-spectrum",
        "type: spectral-model",
        "default-megacomplex: decay",
        "default-megacomplex: spectral",
        "spectral_relations",
        "relations",
        "spectral_constraints",
        "constraints",
        "equal_area_penalties",
        "center_dispersion",
        "width_dispersion",
    ),
)
def test_model_spec_deprecations(
    monkeypatch: MonkeyPatch,
    model_yml_str: str,
    expected_nr_of_warnings: int,
    expected_key: str,
    expected_value: Any,
):
    """Warning gets emitted by load_model"""
    return_dicts = []
    with monkeypatch.context() as m:
        m.setattr(yml_module, "sanitize_yaml", lambda spec: return_dicts.append(spec))
        record, _ = deprecation_warning_on_call_test_helper(
            load_model, args=(model_yml_str,), kwargs={"format_name": "yml_str"}
        )

        return_dict = return_dicts[0]

        assert expected_key in return_dict
        assert return_dict[expected_key] == expected_value

        assert len(record) == expected_nr_of_warnings


@pytest.mark.parametrize(
    "scheme_yml_str, expected_key, expected_value",
    (
        ("maximum-number-function-evaluations: 12", "maximum_number_function_evaluations", 12),
        ("non-negative-least-squares: true", "non_negative_least_squares", True),
    ),
)
def test_scheme_spec_deprecations(
    monkeypatch: MonkeyPatch,
    scheme_yml_str: str,
    expected_key: str,
    expected_value: Any,
):
    """Warning gets emitted by load_model"""
    return_dicts = []
    with monkeypatch.context() as m:
        m.setattr(
            yml_module, "fromdict", lambda _, spec, *args, **kwargs: return_dicts.append(spec)
        )
        record, _ = deprecation_warning_on_call_test_helper(
            load_scheme, args=(scheme_yml_str,), kwargs={"format_name": "yml_str"}
        )

        return_dict = return_dicts[0]

        assert expected_key in return_dict
        assert return_dict[expected_key] == expected_value

        assert len(record) == 1



from __future__ import annotations

import xarray as xr

from glotaran.io import DataIoInterface
from glotaran.io import register_data_io


@register_data_io("nc")
class NetCDFDataIo(DataIoInterface):
    def load_dataset(self, file_name: str) -> xr.Dataset | xr.DataArray:
        with xr.open_dataset(file_name) as ds:
            return ds.load()

    def save_dataset(
        self,
        dataset: xr.Dataset,
        file_name: str,
        *,
        data_filters: list[str] | None = None,
    ):

        data_to_save = dataset if data_filters is None else dataset[data_filters]
        data_to_save.to_netcdf(file_name, mode="w")



"""Helper functions."""
from __future__ import annotations

from typing import List
from typing import Tuple

from glotaran.model.item import model_item


@model_item(
    properties={
        "interval": {"type": List[Tuple[float, float]], "default": None, "allow_none": True},
    },
    has_label=False,
)
class IntervalProperty:
    """Applies a relation between clps as

    :math:`source = parameter * target`.
    """

    def applies(self, value: float) -> bool:
        """
        Returns true if ``value`` is in one of the intervals.

        Parameters
        ----------
        value : float

        Returns
        -------
        applies : bool

        """
        if self.interval is None:
            return True

        def applies(interval):
            return interval[0] <= value <= interval[1]

        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return any([applies(i) for i in self.interval])



"""This package contains the initial concentration item."""
from __future__ import annotations

import copy
from typing import List

import numpy as np

from glotaran.model import model_item
from glotaran.parameter import Parameter


@model_item(
    properties={
        "compartments": List[str],
        "parameters": List[Parameter],
        "exclude_from_normalize": {"type": List[str], "default": []},
    }
)
class InitialConcentration:
    """An initial concentration describes the population of the compartments at
    the beginning of an experiment."""

    def normalized(self) -> InitialConcentration:
        parameters = np.array(self.parameters)
        idx = [c not in self.exclude_from_normalize for c in self.compartments]
        parameters[idx] /= np.sum(parameters[idx])
        new = copy.deepcopy(self)
        for i, value in enumerate(parameters):
            new.parameters[i].value = value
        return new



from copy import deepcopy

import numpy as np
import pytest

from glotaran.analysis.optimization_group import OptimizationGroup
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import TwoCompartmentDecay as suite
from glotaran.model import EqualAreaPenalty
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


@pytest.mark.parametrize("index_dependent", [True, False])
@pytest.mark.parametrize("link_clp", [True, False])
def test_penalties(index_dependent, link_clp):
    model = deepcopy(suite.model)
    model.dataset_group_models["default"].link_clp = link_clp
    model.megacomplex["m1"].is_index_dependent = index_dependent
    model.clp_area_penalties.append(
        EqualAreaPenalty.from_dict(
            {
                "source": "s1",
                "source_intervals": [(1, 20)],
                "target": "s2",
                "target_intervals": [(20, 45)],
                "parameter": "3",
                "weight": 10,
            }
        )
    )
    parameters = ParameterGroup.from_list([11e-4, 22e-5, 2])

    global_axis = np.arange(50)

    print(f"{link_clp=}\n{index_dependent=}")
    dataset = simulate(
        suite.sim_model,
        "dataset1",
        parameters,
        {"global": global_axis, "model": suite.model_axis},
    )
    scheme = Scheme(model=model, parameters=parameters, data={"dataset1": dataset})
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])

    assert isinstance(optimization_group.additional_penalty, np.ndarray)
    assert optimization_group.additional_penalty.size == 1
    assert optimization_group.additional_penalty[0] != 0
    assert isinstance(optimization_group.full_penalty, np.ndarray)
    assert (
        optimization_group.full_penalty.size
        == (suite.model_axis.size * global_axis.size) + optimization_group.additional_penalty.size
    )



"""The Weight property class."""

from typing import List
from typing import Tuple

from glotaran.model.item import model_item


@model_item(
    properties={
        "datasets": {type: List[str]},
        "global_interval": {
            "type": List[Tuple[float, float]],
            "default": None,
            "allow_none": True,
        },
        "model_interval": {
            "type": List[Tuple[float, float]],
            "default": None,
            "allow_none": True,
        },
        "value": {"type": float},
    },
    has_label=False,
)
class Weight:
    """The `Weight` class describes a value by which a dataset will scaled.

    `global_interval` and `model_interval` are optional. The whole range of the dataset
    will be used if not set.
    """



from __future__ import annotations

import warnings
from importlib import import_module
from typing import TYPE_CHECKING

import pytest

from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.deprecation_utils import module_attribute
from glotaran.io import load_dataset
from glotaran.parameter import ParameterGroup
from glotaran.project import Result
from glotaran.project import Scheme
from glotaran.project import result as project_result
from glotaran.project import scheme as project_scheme

if TYPE_CHECKING:
    from _pytest.recwarn import WarningsRecorder


def check_recwarn(records: WarningsRecorder, warn_nr=1):

    for record in records:
        print(record)

    assert len(records) == warn_nr
    assert records[0].category == GlotaranApiDeprecationWarning

    records.clear()


def changed_import_test_warn(
    recwarn: WarningsRecorder, module_qual_name: str, *, attribute_name: str = None, warn_nr=1
):
    """Helper for testing changed imports, returning the imported item.

    Parameters
    ----------
    module_qual_name : str
        Fully qualified name for a module e.g. ``glotaran.model.base_model``
    attribute_name : str, optional
        Name of the attribute e.g. ``Model``

    Returns
    -------
    Any
        Module attribute or module
    """

    warnings.simplefilter("always")

    recwarn.clear()

    if attribute_name is not None:
        result = module_attribute(module_qual_name, attribute_name)
    else:
        result = import_module(module_qual_name)
    check_recwarn(recwarn, warn_nr=warn_nr)
    return result


@pytest.mark.xfail(strict=True, reason="Fail if no warning")
def test_changed_import_test_warn_attribute_no_warn(
    recwarn: WarningsRecorder,
):
    """Module attribute import not warning"""
    changed_import_test_warn(recwarn, "glotaran.parameter", attribute_name="ParameterGroup")


@pytest.mark.xfail(strict=True, reason="Fail if no warning")
def test_changed_import_test_warn_module_no_warn(
    recwarn: WarningsRecorder,
):
    """Module import not warning"""
    changed_import_test_warn(recwarn, "glotaran.parameter")


def test_root_ParameterGroup(recwarn: WarningsRecorder):
    """glotaran.ParameterGroup"""
    result = changed_import_test_warn(recwarn, "glotaran", attribute_name="ParameterGroup")

    assert result == ParameterGroup


def test_analysis_result(recwarn: WarningsRecorder):
    """Usage of glotaran.analysis.result"""
    warnings.simplefilter("always")

    from glotaran.analysis import result

    assert len(recwarn) == 0
    assert result.Result == project_result.Result  # type: ignore [attr-defined]

    check_recwarn(recwarn)


def test_analysis_result_from_import(recwarn: WarningsRecorder):
    """Same as 'from glotaran.analysis.result import Result' as analysis_result"""
    analysis_result = changed_import_test_warn(
        recwarn, "glotaran.analysis.result", attribute_name="Result"
    )

    assert analysis_result == Result


def test_analysis_scheme(recwarn: WarningsRecorder):
    """Usage of glotaran.analysis.scheme"""
    warnings.simplefilter("always")

    from glotaran.analysis import scheme as analysis_scheme

    assert len(recwarn) == 0
    assert analysis_scheme.Scheme == project_scheme.Scheme  # type: ignore [attr-defined]

    check_recwarn(recwarn)


def test_analysis_scheme_from_import(recwarn: WarningsRecorder):
    """Same as 'from glotaran.analysis.scheme import Scheme as analysis_scheme'"""
    analysis_scheme = changed_import_test_warn(
        recwarn, "glotaran.analysis.scheme", attribute_name="Scheme"
    )

    assert analysis_scheme == Scheme


def test_io_read_data_file(recwarn: WarningsRecorder):
    """glotaran.io.read_data_file"""
    result = changed_import_test_warn(recwarn, "glotaran.io", attribute_name="read_data_file")

    assert result.__code__ == load_dataset.__code__



"""Utility functions for io plugin."""

from __future__ import annotations

import os
from functools import partial
from functools import wraps
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import TypeVar
from typing import cast

DecoratedFunc = TypeVar("DecoratedFunc", bound=Callable[..., Any])  # decorated function


def inferr_file_format(
    file_path: str | os.PathLike[str], *, needs_to_exist: bool = True, allow_folder=False
) -> str:
    """Inferr format of a file if it exists.

    Parameters
    ----------
    file_path : str
        Path/str to the file.
    needs_to_exist : bool
        Whether or not a file need to exists for an successful format inferring.
        While write functions don't need the file to exists, load functions do.
    allow_folder: bool
        Whether or not to allow the format to be ``folder``.
        This is only used in ``save_result``.

    Returns
    -------
    str
        File extension without the leading dot.

    Raises
    ------
    ValueError
        If file doesn't exists.
    ValueError
        If file has no extension.
    """
    if not os.path.isfile(file_path) and needs_to_exist and not allow_folder:
        raise ValueError(f"There is no file {file_path!r}.")

    _, file_format = os.path.splitext(file_path)
    if file_format == "":
        if allow_folder:
            return "folder"
        else:
            raise ValueError(
                f"Cannot determine format of file {file_path!r}, "
                "please provide an explicit format."
            )
    else:
        return file_format.lstrip(".")


def not_implemented_to_value_error(func: DecoratedFunc) -> DecoratedFunc:
    """Decorate a function to raise ValueError instead of NotImplementedError.

    This decorator is supposed to be used on functions which call functions
    that might raise a NotImplementedError, but raise ValueError instead with
    the same error text.

    Parameters
    ----------
    func : DecoratedFunc
        Function to be decorated.

    Returns
    -------
    DecoratedFunc
        Wrapped function.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except NotImplementedError as error:
            raise ValueError(error.args)

    return cast(DecoratedFunc, wrapper)


def protect_from_overwrite(path: str | os.PathLike[str], *, allow_overwrite: bool = False) -> None:
    """Raise FileExistsError if files already exists and allow_overwrite isn't True.

    Parameters
    ----------
    path : str
        Path to a file or folder.
    allow_overwrite : bool
        Whether or not to allow overwriting existing files, by default False

    Raises
    ------
    FileExistsError
        If path points to an existing file.
    FileExistsError
        If path points to an existing folder which is not empty.
    """
    user_info = (
        "To protect users overwriting existing files is deactivated by default. "
        "If you are absolutely sure this is what you want and need to do you can "
        "use the argument 'allow_overwrite=True'."
    )
    if allow_overwrite:
        return
    elif os.path.isfile(path):
        raise FileExistsError(f"The file {path!r} already exists. \n{user_info}")
    elif os.path.isdir(path) and os.listdir(str(path)):
        raise FileExistsError(
            f"The folder {path!r} already exists and is not empty. \n{user_info}"
        )


def bool_str_repr(value: Any, true_repr: str = "*", false_repr: str = "/") -> Any:
    """Replace boolean value with string repr.

    This function is a helper for table representation (e.g. with tabulate)
    of boolean values.

    Parameters
    ----------
    value : Any
        Arbitrary value
    true_repr : str
        Desired repr for ``True``, by default "*"
    false_repr : str
        Desired repr for ``False``, by default "/"

    Returns
    -------
    Any
        Original value or desired repr for bool

    Examples
    --------
    >>> table_data = [["foo", True, False], ["bar", False, True]]
    >>> print(tabulate(map(lambda x: map(bool_table_repr, x), table_data)))
    ---  -  -
    foo  *  /
    bar  /  *
    ---  -  -
    """
    # since bool is a subclass of int we can't use isinstance
    if type(value) == bool:
        if value:
            return true_repr
        else:
            return false_repr
    else:
        return value


def bool_table_repr(
    table_data: Iterable[Iterable[Any]], true_repr: str = "*", false_repr: str = "/"
) -> Iterator[Iterator[Any]]:
    """Replace boolean value with string repr for all table values.

    This function is an implementation of :func:`bool_str_repr` for a
    2D table, for easy usage with tabulate.

    Parameters
    ----------
    table_data : Iterable[Iterable[Any]]
        Data of the table e.g. a list of lists.
    true_repr : str
        Desired repr for ``True``, by default "*"
    false_repr : str
        Desired repr for ``False``, by default "/"

    Returns
    -------
    Iterator[Iterator[Any]]
        ``table_data`` with original values or desired repr for bool

    See Also
    --------
    bool_str_repr

    Examples
    --------
    >>> table_data = [["foo", True, False], ["bar", False, True]]
    >>> print(tabulate(bool_table_repr(table_data))
    ---  -  -
    foo  *  /
    bar  /  *
    ---  -  -
    """
    bool_repr = partial(bool_str_repr, true_repr=true_repr, false_repr=false_repr)
    return map(lambda value: map(bool_repr, value), table_data)



"""This package contains functions for model simulation and fitting."""
from glotaran.deprecation.deprecation_utils import deprecate_submodule

result = deprecate_submodule(
    deprecated_module_name="glotaran.analysis.result",
    new_module_name="glotaran.project.result",
    to_be_removed_in_version="0.6.0",
)

scheme = deprecate_submodule(
    deprecated_module_name="glotaran.analysis.scheme",
    new_module_name="glotaran.project.scheme",
    to_be_removed_in_version="0.6.0",
)



"""This package contains compartment constraint items."""
from __future__ import annotations

from glotaran.model.interval_property import IntervalProperty
from glotaran.model.item import model_item
from glotaran.model.item import model_item_typed


@model_item(
    properties={
        "target": str,
    },
    has_type=True,
    has_label=False,
)
class OnlyConstraint(IntervalProperty):
    """A only constraint sets the calculated matrix row of a compartment to 0
    outside the given intervals."""

    def applies(self, value: float) -> bool:
        """
        Returns true if ``value`` is in one of the intervals.

        Parameters
        ----------
        index : float

        Returns
        -------
        applies : bool

        """
        return not super().applies(value)


@model_item(
    properties={
        "target": str,
    },
    has_type=True,
    has_label=False,
)
class ZeroConstraint(IntervalProperty):
    """A zero constraint sets the calculated matrix row of a compartment to 0
    in the given intervals."""


@model_item_typed(
    types={
        "only": OnlyConstraint,
        "zero": ZeroConstraint,
    },
    has_label=False,
)
class Constraint:
    """A constraint is applied on one clp on one or many
    intervals on the estimated axis type.

    There are two types: zero and equal. See the documentation of
    the respective classes for details.
    """



"""The DatasetModel class."""
from __future__ import annotations

from collections import Counter
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.model.item import model_item
from glotaran.model.item import model_item_validator

if TYPE_CHECKING:
    from typing import Any
    from typing import Generator
    from typing import Hashable

    from glotaran.model.megacomplex import Megacomplex
    from glotaran.model.model import Model
    from glotaran.parameter import Parameter


def create_dataset_model_type(properties: dict[str, Any]) -> type[DatasetModel]:
    @model_item(properties=properties)
    class ModelDatasetModel(DatasetModel):
        pass

    return ModelDatasetModel


class DatasetModel:
    """A `DatasetModel` describes a dataset in terms of a glotaran model.
    It contains references to model items which describe the physical model for
    a given dataset.

    A general dataset descriptor assigns one or more megacomplexes and a scale
    parameter.
    """

    def iterate_megacomplexes(
        self,
    ) -> Generator[tuple[Parameter | int | None, Megacomplex | str], None, None]:
        """Iterates of der dataset model's megacomplexes."""
        for i, megacomplex in enumerate(self.megacomplex):
            scale = self.megacomplex_scale[i] if self.megacomplex_scale is not None else None
            yield scale, megacomplex

    def iterate_global_megacomplexes(
        self,
    ) -> Generator[tuple[Parameter | int | None, Megacomplex | str], None, None]:
        """Iterates of der dataset model's global megacomplexes."""
        for i, megacomplex in enumerate(self.global_megacomplex):
            scale = (
                self.global_megacomplex_scale[i]
                if self.global_megacomplex_scale is not None
                else None
            )
            yield scale, megacomplex

    def get_model_dimension(self) -> str:
        """Returns the dataset model's model dimension."""
        if not hasattr(self, "_model_dimension"):
            if len(self.megacomplex) == 0:
                raise ValueError(f"No megacomplex set for dataset model '{self.label}'")
            if isinstance(self.megacomplex[0], str):
                raise ValueError(f"Dataset model '{self.label}' was not filled")
            self._model_dimension = self.megacomplex[0].dimension
            if any(self._model_dimension != m.dimension for m in self.megacomplex):
                raise ValueError(
                    f"Megacomplex dimensions do not match for dataset model '{self.label}'."
                )
        return self._model_dimension

    def finalize_data(self, dataset: xr.Dataset) -> None:
        is_full_model = self.has_global_model()
        for megacomplex in self.megacomplex:
            megacomplex.finalize_data(self, dataset, is_full_model=is_full_model)
        if is_full_model:
            for megacomplex in self.global_megacomplex:
                megacomplex.finalize_data(
                    self, dataset, is_full_model=is_full_model, as_global=True
                )

    def overwrite_model_dimension(self, model_dimension: str) -> None:
        """Overwrites the dataset model's model dimension."""
        self._model_dimension = model_dimension

    # TODO: make explicit we only support 2 dimensions at present
    # TODO: the global dimension should become a flexible index (MultiIndex)
    # the user can then specify the name of the MultiIndex global dimension
    # using the function overwrite_global_dimension
    # e.g. in FLIM, x, y dimension may get 'flattened' to a MultiIndex 'pixel'
    def get_global_dimension(self) -> str:
        """Returns the dataset model's global dimension."""
        if not hasattr(self, "_global_dimension"):
            if self.has_global_model():
                if isinstance(self.global_megacomplex[0], str):
                    raise ValueError(f"Dataset model '{self.label}' was not filled")
                self._global_dimension = self.global_megacomplex[0].dimension
                if any(self._global_dimension != m.dimension for m in self.global_megacomplex):
                    raise ValueError(
                        "Global megacomplex dimensions do not "
                        f"match for dataset model '{self.label}'."
                    )
            elif hasattr(self, "_coords"):
                return next(dim for dim in self._coords if dim != self.get_model_dimension())
            else:
                if not hasattr(self, "_data"):
                    raise ValueError(f"Data not set for dataset model '{self.label}'")
                self._global_dimension = next(
                    dim for dim in self._data.data.dims if dim != self.get_model_dimension()
                )
        return self._global_dimension

    def overwrite_global_dimension(self, global_dimension: str) -> None:
        """Overwrites the dataset model's global dimension."""
        self._global_dimension = global_dimension

    def swap_dimensions(self) -> None:
        """Swaps the dataset model's global and model dimension."""
        global_dimension = self.get_model_dimension()
        model_dimension = self.get_global_dimension()
        self.overwrite_global_dimension(global_dimension)
        self.overwrite_model_dimension(model_dimension)

    def set_data(self, dataset: xr.Dataset) -> DatasetModel:
        """Sets the dataset model's data."""
        self._coords = {name: dim.values for name, dim in dataset.coords.items()}
        self._data: np.ndarray = dataset.data.values
        self._weight: np.ndarray | None = dataset.weight.values if "weight" in dataset else None
        if self._weight is not None:
            self._data = self._data * self._weight
        return self

    def get_data(self) -> np.ndarray:
        """Gets the dataset model's data."""
        return self._data

    def get_weight(self) -> np.ndarray | None:
        """Gets the dataset model's weight."""
        return self._weight

    def is_index_dependent(self) -> bool:
        """Indicates if the dataset model is index dependent."""
        if hasattr(self, "_index_dependent"):
            return self._index_dependent
        return any(m.index_dependent(self) for m in self.megacomplex)

    def overwrite_index_dependent(self, index_dependent: bool):
        """Overrides the index dependency of the dataset"""
        self._index_dependent = index_dependent

    def has_global_model(self) -> bool:
        """Indicates if the dataset model can model the global dimension."""
        return self.global_megacomplex is not None and len(self.global_megacomplex) != 0

    def set_coordinates(self, coords: dict[str, np.ndarray]):
        """Sets the dataset model's coordinates."""
        self._coords = coords

    def get_coordinates(self) -> dict[Hashable, np.ndarray]:
        """Gets the dataset model's coordinates."""
        return self._coords

    def get_model_axis(self) -> np.ndarray:
        """Gets the dataset model's model axis."""
        return self._coords[self.get_model_dimension()]

    def get_global_axis(self) -> np.ndarray:
        """Gets the dataset model's global axis."""
        return self._coords[self.get_global_dimension()]

    @model_item_validator(False)
    def ensure_unique_megacomplexes(self, model: Model) -> list[str]:
        """Ensure that unique megacomplexes Are only used once per dataset.

        Parameters
        ----------
        model : Model
            Model object using this dataset model.

        Returns
        -------
        list[str]
            Error messages to be shown when the model gets validated.
        """
        glotaran_unique_megacomplex_types = []

        for megacomplex_name in self.megacomplex:
            try:
                megacomplex_instance = model.megacomplex[megacomplex_name]
                if type(megacomplex_instance).glotaran_unique() is True:
                    type_name = megacomplex_instance.type or megacomplex_instance.name
                    glotaran_unique_megacomplex_types.append(type_name)
            except KeyError:
                pass

        return [
            f"Multiple instances of unique megacomplex type {type_name!r} "
            f"in dataset {self.label!r}"
            for type_name, count in Counter(glotaran_unique_megacomplex_types).most_common()
            if count > 1
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.analysis.optimization_group_calculator import OptimizationGroupCalculator
from glotaran.analysis.util import CalculatedMatrix
from glotaran.analysis.util import apply_weight
from glotaran.analysis.util import calculate_clp_penalties
from glotaran.analysis.util import calculate_matrix
from glotaran.analysis.util import reduce_matrix
from glotaran.analysis.util import retrieve_clps
from glotaran.model import DatasetModel

if TYPE_CHECKING:
    from glotaran.analysis.optimization_group import OptimizationGroup


class OptimizationGroupCalculatorUnlinked(OptimizationGroupCalculator):
    """Represents a problem where the clps are not linked."""

    def __init__(self, group: OptimizationGroup):
        super().__init__(group)

        self._global_matrices = {}
        self._flattened_data = {}
        self._flattened_weights = {}
        for label, dataset_model in group.dataset_models.items():
            if dataset_model.has_global_model():
                self._flattened_data[label] = dataset_model.get_data().T.flatten()
                weight = dataset_model.get_weight()
                if weight is not None:
                    weight = weight.T.flatten()
                    self._flattened_data[label] *= weight
                    self._flattened_weight[label] = weight

    @property
    def global_matrices(self) -> dict[str, CalculatedMatrix]:
        return self._global_matrices

    def calculate_matrices(
        self,
    ) -> tuple[
        dict[str, CalculatedMatrix | list[CalculatedMatrix]],
        dict[str, CalculatedMatrix | list[CalculatedMatrix]],
    ]:
        """Calculates the model matrices."""

        self._group._matrices = {}
        self._global_matrices = {}
        self._group._reduced_matrices = {}

        for label, dataset_model in self._group.dataset_models.items():

            if dataset_model.is_index_dependent():
                self._calculate_index_dependent_matrix(label, dataset_model)
            else:
                self._calculate_index_independent_matrix(label, dataset_model)

            if dataset_model.has_global_model():
                self._calculate_global_matrix(label, dataset_model)

        return self._group._matrices, self._group._reduced_matrices

    def _calculate_index_dependent_matrix(self, label: str, dataset_model: DatasetModel):
        self._group._matrices[label] = []
        self._group._reduced_matrices[label] = []
        for i, index in enumerate(dataset_model.get_global_axis()):
            matrix = calculate_matrix(
                dataset_model,
                {dataset_model.get_global_dimension(): i},
            )
            self._group._matrices[label].append(matrix)
            if not dataset_model.has_global_model():
                reduced_matrix = reduce_matrix(
                    matrix, self._group.model, self._group.parameters, index
                )
                self._group._reduced_matrices[label].append(reduced_matrix)

    def _calculate_index_independent_matrix(self, label: str, dataset_model: DatasetModel):
        matrix = calculate_matrix(dataset_model, {})
        self._group._matrices[label] = matrix
        if not dataset_model.has_global_model():
            reduced_matrix = reduce_matrix(matrix, self._group.model, self._group.parameters, None)
            self._group._reduced_matrices[label] = reduced_matrix

    def _calculate_global_matrix(self, label: str, dataset_model: DatasetModel):
        matrix = calculate_matrix(dataset_model, {}, as_global_model=True)
        self._global_matrices[label] = matrix

    def calculate_residual(
        self,
    ) -> tuple[
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the residuals."""

        self._group._reduced_clps = {}
        self._group._clps = {}
        self._group._weighted_residuals = {}
        self._group._residuals = {}
        self._group._additional_penalty = []

        for label, dataset_model in self._group._dataset_models.items():
            if dataset_model.has_global_model():
                self._calculate_full_model_residual(label, dataset_model)
            else:
                self._calculate_residual(label, dataset_model)

        self._group._additional_penalty = (
            np.concatenate(self._group._additional_penalty)
            if len(self._group._additional_penalty) != 0
            else []
        )
        return (
            self._group._reduced_clps,
            self._group._clps,
            self._group._weighted_residuals,
            self._group._residuals,
        )

    def _calculate_residual(self, label: str, dataset_model: DatasetModel):
        self._group._reduced_clps[label] = []
        self._group._clps[label] = []
        self._group._weighted_residuals[label] = []
        self._group._residuals[label] = []

        data = dataset_model.get_data()
        global_axis = dataset_model.get_global_axis()

        for i, index in enumerate(global_axis):
            reduced_clp_labels, reduced_matrix = (
                self._group.reduced_matrices[label][i]
                if dataset_model.is_index_dependent()
                else self._group.reduced_matrices[label]
            )
            if not dataset_model.is_index_dependent():
                reduced_matrix = reduced_matrix.copy()

            if dataset_model.scale is not None:
                reduced_matrix *= dataset_model.scale

            weight = dataset_model.get_weight()
            if weight is not None:
                apply_weight(reduced_matrix, weight[:, i])

            reduced_clps, residual = self._group._residual_function(reduced_matrix, data[:, i])

            self._group._reduced_clps[label].append(reduced_clps)

            clp_labels = self._get_clp_labels(label, i)
            self._group._clps[label].append(
                retrieve_clps(
                    self._group.model,
                    self._group.parameters,
                    clp_labels,
                    reduced_clp_labels,
                    reduced_clps,
                    index,
                )
            )
            self._group._weighted_residuals[label].append(residual)
            if weight is not None:
                self._group._residuals[label].append(residual / weight[:, i])
            else:
                self._group._residuals[label].append(residual)

        clp_labels = self._get_clp_labels(label)
        additional_penalty = calculate_clp_penalties(
            self._group.model,
            self._group.parameters,
            clp_labels,
            self._group._clps[label],
            global_axis,
            self._group.dataset_models,
        )
        if additional_penalty.size != 0:
            self._group._additional_penalty.append(additional_penalty)

    def _calculate_full_model_residual(self, label: str, dataset_model: DatasetModel):

        model_matrix = self._group.matrices[label]
        global_matrix = self.global_matrices[label].matrix

        if dataset_model.is_index_dependent():
            matrix = np.concatenate(
                [
                    np.kron(global_matrix[i, :], model_matrix[i].matrix)
                    for i in range(global_matrix.shape[0])
                ]
            )
        else:
            matrix = np.kron(global_matrix, model_matrix.matrix)
        weight = self._flattened_weights.get(label)
        if weight is not None:
            apply_weight(matrix, weight)
        data = self._flattened_data[label]
        (
            self._group._clps[label],
            self._group._weighted_residuals[label],
        ) = self._group._residual_function(matrix, data)

        self._group._residuals[label] = self._group._weighted_residuals[label]
        if weight is not None:
            self._group._residuals[label] /= weight

    def _get_clp_labels(self, label: str, index: int = 0):
        return (
            self._group.matrices[label][index].clp_labels
            if self._group.dataset_models[label].is_index_dependent()
            else self._group.matrices[label].clp_labels
        )

    def create_index_dependent_result_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        """Creates a result datasets for index dependent matrices."""

        model_dimension = self._group.dataset_models[label].get_model_dimension()
        global_dimension = self._group.dataset_models[label].get_global_dimension()

        dataset.coords["clp_label"] = self._get_clp_labels(label)
        dataset["matrix"] = (
            (
                (global_dimension),
                (model_dimension),
                ("clp_label"),
            ),
            np.asarray([m.matrix for m in self._group.matrices[label]]),
        )

        if self._group.dataset_models[label].has_global_model():
            self._add_global_matrix_to_dataset(label, dataset)
            self._add_full_model_residual_and_clp_to_dataset(label, dataset)
        else:
            self._add_residual_and_clp_to_dataset(label, dataset)

        return dataset

    def create_index_independent_result_dataset(
        self, label: str, dataset: xr.Dataset
    ) -> xr.Dataset:
        """Creates a result datasets for index independent matrices."""

        matrix = self._group.matrices[label]
        dataset.coords["clp_label"] = matrix.clp_labels
        model_dimension = self._group.dataset_models[label].get_model_dimension()
        dataset["matrix"] = (
            (
                (model_dimension),
                ("clp_label"),
            ),
            matrix.matrix,
        )

        if self._group.dataset_models[label].has_global_model():
            self._add_global_matrix_to_dataset(label, dataset)
            self._add_full_model_residual_and_clp_to_dataset(label, dataset)
        else:
            self._add_residual_and_clp_to_dataset(label, dataset)

        return dataset

    def _add_global_matrix_to_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        matrix = self.global_matrices[label]
        dataset.coords["global_clp_label"] = matrix.clp_labels
        global_dimension = self._group.dataset_models[label].get_global_dimension()
        dataset["global_matrix"] = (
            (
                (global_dimension),
                ("global_clp_label"),
            ),
            matrix.matrix,
        )

    def _add_residual_and_clp_to_dataset(self, label: str, dataset: xr.Dataset):
        model_dimension = self._group.dataset_models[label].get_model_dimension()
        global_dimension = self._group.dataset_models[label].get_global_dimension()
        dataset["clp"] = (
            (
                (global_dimension),
                ("clp_label"),
            ),
            np.asarray(self._group.clps[label]),
        )
        dataset["weighted_residual"] = (
            (
                (model_dimension),
                (global_dimension),
            ),
            np.transpose(np.asarray(self._group.weighted_residuals[label])),
        )
        dataset["residual"] = (
            (
                (model_dimension),
                (global_dimension),
            ),
            np.transpose(np.asarray(self._group.residuals[label])),
        )

    def _add_full_model_residual_and_clp_to_dataset(self, label: str, dataset: xr.Dataset):
        model_dimension = self._group.dataset_models[label].get_model_dimension()
        global_dimension = self._group.dataset_models[label].get_global_dimension()
        dataset["clp"] = (
            (
                ("global_clp_label"),
                ("clp_label"),
            ),
            self._group.clps[label].reshape(
                (dataset.coords["global_clp_label"].size, dataset.coords["clp_label"].size)
            ),
        )
        dataset["weighted_residual"] = (
            (
                (model_dimension),
                (global_dimension),
            ),
            self._group.weighted_residuals[label].T.reshape(dataset.data.shape),
        )
        dataset["residual"] = (
            (
                (model_dimension),
                (global_dimension),
            ),
            self._group.residuals[label].T.reshape(dataset.data.shape),
        )

    def calculate_full_penalty(self) -> np.ndarray:
        if self._group._full_penalty is None:
            residuals = self._group.weighted_residuals
            additional_penalty = self._group.additional_penalty
            residuals = [
                np.concatenate(residuals[label])
                if isinstance(residuals[label], list)
                else residuals[label]
                for label in residuals.keys()
            ]

            self._group._full_penalty = (
                np.concatenate((np.concatenate(residuals), additional_penalty))
                if additional_penalty is not None
                else np.concatenate(residuals)
            )
        return self._group._full_penalty



"""The parameter group class."""

from __future__ import annotations

from copy import copy
from textwrap import indent
from typing import TYPE_CHECKING
from typing import Generator

import asteval
import numpy as np
import pandas as pd
from tabulate import tabulate

from glotaran.deprecation import deprecate
from glotaran.io import save_parameters
from glotaran.parameter.parameter import Parameter
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from glotaran.parameter.parameter_history import ParameterHistory


class ParameterNotFoundException(Exception):
    """Raised when a Parameter is not found in the Group."""

    def __init__(self, path, label):  # noqa: D107
        super().__init__(f"Cannot find parameter {'.'.join(path+[label])!r}")


class ParameterGroup(dict):
    """Represents are group of parameters.

    Can contain other groups, creating a tree-like hierarchy.
    """

    def __init__(self, label: str = None, root_group: ParameterGroup = None):
        """Initialize a :class:`ParameterGroup` instance with ``label``.

        Parameters
        ----------
        label : str
            The label of the group.
        root_group : ParameterGroup
            The root group

        Raises
        ------
        ValueError
            Raised if the an invalid label is given.
        """
        if label is not None and not Parameter.valid_label(label):
            raise ValueError(f"'{label}' is not a valid group label.")
        self._label = label
        self._parameters: dict[str, Parameter] = {}
        self._root_group = root_group
        self._evaluator = (
            asteval.Interpreter(symtable=asteval.make_symbol_table(group=self))
            if root_group is None
            else None
        )
        super().__init__()

    @classmethod
    def from_dict(
        cls,
        parameter_dict: dict[str, dict | list],
        label: str = None,
        root_group: ParameterGroup = None,
    ) -> ParameterGroup:
        """Create a :class:`ParameterGroup` from a dictionary.

        Parameters
        ----------
        parameter_dict : dict[str, dict | list]
            A parameter dictionary containing parameters.
        label : str
            The label of the group.
        root_group : ParameterGroup
            The root group

        Returns
        -------
        ParameterGroup
            The created :class:`ParameterGroup`
        """
        root = cls(label=label, root_group=root_group)
        for label, item in parameter_dict.items():
            label = str(label)
            if isinstance(item, dict):
                root.add_group(cls.from_dict(item, label=label, root_group=root))
            if isinstance(item, list):
                root.add_group(cls.from_list(item, label=label, root_group=root))
        if root_group is None:
            root.update_parameter_expression()
        return root

    @classmethod
    def from_list(
        cls,
        parameter_list: list[float | list],
        label: str = None,
        root_group: ParameterGroup = None,
    ) -> ParameterGroup:
        """Create a :class:`ParameterGroup` from a list.

        Parameters
        ----------
        parameter_list : list[float | list]
            A parameter list containing parameters
        label : str
            The label of the group.
        root_group : ParameterGroup
            The root group

        Returns
        -------
        ParameterGroup
            The created :class:`ParameterGroup`
        """
        root = cls(label=label, root_group=root_group)

        # get defaults
        defaults = None
        for item in parameter_list:
            if isinstance(item, dict):
                defaults = item
                break

        for i, item in enumerate(parameter_list):
            if isinstance(item, (str, int, float)):
                try:
                    item = float(item)
                except Exception:
                    pass
            if isinstance(item, (float, int, list)):
                root.add_parameter(
                    Parameter.from_list_or_value(item, label=str(i + 1), default_options=defaults)
                )
        if root_group is None:
            root.update_parameter_expression()
        return root

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, source: str = "DataFrame") -> ParameterGroup:
        """Create a :class:`ParameterGroup` from a :class:`pandas.DataFrame`.

        Parameters
        ----------
        df : pd.DataFrame
            The source data frame.
        source : str
            Optional name of the source file, used for error messages.

        Returns
        -------
        ParameterGroup
            The created parameter group.

        Raises
        ------
        ValueError
            Raised if the columns 'label' or 'value' doesn't exist. Also raised if the columns
            'minimum', 'maximum' or 'values' contain non numeric values or if the columns
            'non-negative' or 'vary' are no boolean.
        """
        for column_name in ["label", "value"]:
            if column_name not in df:
                raise ValueError(f"Missing column '{column_name}' in '{source}'")

        for column_name in ["minimum", "maximum", "value"]:
            if column_name in df and any(not np.isreal(v) for v in df[column_name]):
                raise ValueError(f"Column '{column_name}' in '{source}' has non numeric values")

        for column_name in ["non-negative", "vary"]:
            if column_name in df and any(not isinstance(v, bool) for v in df[column_name]):
                raise ValueError(f"Column '{column_name}' in '{source}' has non boolean values")

        root = cls()

        for i, full_label in enumerate(df["label"]):
            path = full_label.split(".")
            group = root
            while len(path) > 1:
                group_label = path.pop(0)
                if group_label not in group:
                    group.add_group(ParameterGroup(label=group_label, root_group=group))
                group = group[group_label]
            label = path.pop()
            value = df["value"][i]
            minimum = df["minimum"][i] if "minimum" in df else -np.inf
            maximum = df["maximum"][i] if "maximum" in df else np.inf
            non_negative = df["non-negative"][i] if "non-negative" in df else False
            vary = df["vary"][i] if "vary" in df else True
            expression = (
                df["expression"][i]
                if "expression" in df and isinstance(df["expression"][i], str)
                else None
            )

            parameter = Parameter(
                label=label,
                full_label=full_label,
                value=value,
                expression=expression,
                maximum=maximum,
                minimum=minimum,
                non_negative=non_negative,
                vary=vary,
            )
            group.add_parameter(parameter)
        root.update_parameter_expression()
        return root

    @property
    def label(self) -> str | None:
        """Label of the group.

        Returns
        -------
        str
            The label of the group.
        """
        return self._label

    @property
    def root_group(self) -> ParameterGroup | None:
        """Root of the group.

        Returns
        -------
        ParameterGroup
            The root group.
        """
        return self._root_group

    def to_dataframe(self) -> pd.DataFrame:
        """Create a pandas data frame from the group.

        Returns
        -------
        pd.DataFrame
            The created data frame.
        """
        parameter_dict: dict[str, list[str | float | bool | None]] = {
            "label": [],
            "value": [],
            "minimum": [],
            "maximum": [],
            "vary": [],
            "non-negative": [],
            "expression": [],
        }
        for label, parameter in self.all():
            parameter_dict["label"].append(label)
            parameter_dict["value"].append(parameter.value)
            parameter_dict["minimum"].append(parameter.minimum)
            parameter_dict["maximum"].append(parameter.maximum)
            parameter_dict["vary"].append(parameter.vary)
            parameter_dict["non-negative"].append(parameter.non_negative)
            parameter_dict["expression"].append(parameter.expression)
        return pd.DataFrame(parameter_dict)

    @deprecate(
        deprecated_qual_name_usage=(
            "glotaran.parameter.ParameterGroup.to_csv(file_name=<parameters.csv>)"
        ),
        new_qual_name_usage=(
            "glotaran.io.save_parameters(parameters, "
            'file_name=<parameters.csv>, format_name="csv")'
        ),
        to_be_removed_in_version="0.7.0",
        importable_indices=(2, 1),
    )
    def to_csv(self, filename: str, delimiter: str = ",") -> None:
        """Save a :class:`ParameterGroup` to a CSV file.

        Warning
        -------
        Deprecated use
        ``glotaran.io.save_parameters(parameters, file_name=<parameters.csv>, format_name="csv")``
        instead.

        Parameters
        ----------
        filename : str
            File to write the parameter specs to.
        delimiter : str
            Character to separate columns., by default ","
        """
        save_parameters(self, file_name=filename, allow_overwrite=True, sep=delimiter)

    def add_parameter(self, parameter: Parameter | list[Parameter]):
        """Add a :class:`Parameter` to the group.

        Parameters
        ----------
        parameter : Parameter | list[Parameter]
            The parameter to add.

        Raises
        ------
        TypeError
            If ``parameter`` or any item of it is not an instance of :class:`Parameter`.
        """
        if not isinstance(parameter, list):
            parameter = [parameter]
        if any(not isinstance(p, Parameter) for p in parameter):
            raise TypeError("Parameter must be  instance of glotaran.parameter.Parameter")
        for p in parameter:
            p.index = len(self._parameters) + 1
            if p.label is None:
                p.label = f"{p.index}"
            p.full_label = f"{self.label}.{p.label}" if self.label else p.label
            self._parameters[p.label] = p

    def add_group(self, group: ParameterGroup):
        """Add a :class:`ParameterGroup` to the group.

        Parameters
        ----------
        group : ParameterGroup
            The group to add.

        Raises
        ------
        TypeError
            Raised if the group is not an instance of :class:`ParameterGroup`.
        """
        if not isinstance(group, ParameterGroup):
            raise TypeError("Group must be glotaran.parameter.ParameterGroup")
        self[group.label] = group

    def get_nr_roots(self) -> int:
        """Return the number of roots of the group.

        Returns
        -------
        int
            The number of roots.
        """
        n = 0
        root = self.root_group
        while root is not None:
            n += 1
            root = root.root_group
        return n

    def groups(self) -> Generator[ParameterGroup, None, None]:
        """Return a generator over all groups and their subgroups.

        Yields
        ------
        ParameterGroup
            A subgroup of :class:`ParameterGroup`.
        """
        for group in self:
            yield from group.groups()

    def has(self, label: str) -> bool:
        """Check if a parameter with the given label is in the group or in a subgroup.

        Parameters
        ----------
        label : str
            The label of the parameter, with its path in a :class:`ParameterGroup` prepended.

        Returns
        -------
        bool
            Whether a parameter with the given label exists in the group.
        """
        try:
            self.get(label)
            return True
        except Exception:
            return False

    def get(self, label: str) -> Parameter:  # type:ignore[override]
        """Get a :class:`Parameter` by its label.

        Parameters
        ----------
        label : str
            The label of the parameter, with its path in a :class:`ParameterGroup` prepended.

        Returns
        -------
        Parameter
            The parameter.

        Raises
        ------
        ParameterNotFoundException
            Raised if no parameter with the given label exists.
        """
        # sometimes the spec parser delivers the labels as int
        label = str(label)

        path = label.split(".")
        label = path.pop()

        # TODO: audit this code
        group = self
        for element in path:
            try:
                group = group[element]
            except KeyError:
                raise ParameterNotFoundException(path, label)
        try:
            return group._parameters[label]
        except KeyError:
            raise ParameterNotFoundException(path, label)

    def copy(self) -> ParameterGroup:
        """Create a copy of the :class:`ParameterGroup`.

        Returns
        -------
        ParameterGroup :
            A copy of the :class:`ParameterGroup`.

        """
        root = ParameterGroup(label=self.label, root_group=self.root_group)

        for label, parameter in self._parameters.items():
            root._parameters[label] = copy(parameter)

        for label, group in self.items():
            root[label] = group.copy()

        return root

    def all(
        self, root: str | None = None, separator: str = "."
    ) -> Generator[tuple[str, Parameter], None, None]:
        """Iterate over all parameter in the group and it's subgroups together with their labels.

        Parameters
        ----------
        root : str
            The label of the root group
        separator : str
            The separator for the parameter labels.

        Yields
        ------
        tuple[str, Parameter]
            A tuple containing the full label of the parameter and the parameter itself.
        """
        root = f"{root}{self.label}{separator}" if root is not None else ""
        for label, p in self._parameters.items():
            yield (f"{root}{label}", p)
        for _, l in self.items():
            yield from l.all(root=root, separator=separator)

    def get_label_value_and_bounds_arrays(
        self, exclude_non_vary: bool = False
    ) -> tuple[list[str], np.ndarray, np.ndarray, np.ndarray]:
        """Return a arrays of all parameter labels, values and bounds.

        Parameters
        ----------
        exclude_non_vary: bool
            If true, parameters with `vary=False` are excluded.

        Returns
        -------
        tuple[list[str], np.ndarray, np.ndarray, np.ndarray]
            A tuple containing a list of parameter labels and
            an array of the values, lower and upper bounds.
        """
        self.update_parameter_expression()

        labels = []
        values = []
        lower_bounds = []
        upper_bounds = []

        for label, parameter in self.all():
            if not exclude_non_vary or parameter.vary:
                labels.append(label)
                value, minimum, maximum = parameter.get_value_and_bounds_for_optimization()
                values.append(value)
                lower_bounds.append(minimum)
                upper_bounds.append(maximum)

        return labels, np.asarray(values), np.asarray(lower_bounds), np.asarray(upper_bounds)

    def set_from_label_and_value_arrays(self, labels: list[str], values: np.ndarray):
        """Update the parameter values from a list of labels and values.

        Parameters
        ----------
        labels : list[str]
            A list of parameter labels.
        values : np.ndarray
            An array of parameter values.

        Raises
        ------
        ValueError
            Raised if the size of the labels does not match the stize of values.
        """
        if len(labels) != len(values):
            raise ValueError(
                f"Length of labels({len(labels)}) not equal to length of values({len(values)})."
            )

        for label, value in zip(labels, values):
            self.get(label).set_value_from_optimization(value)

        self.update_parameter_expression()

    def set_from_history(self, history: ParameterHistory, index: int):
        """Update the :class:`ParameterGroup` with values from a parameter history.

        Parameters
        ----------
        history : ParameterHistory
            The parameter history.
        index : int
            The history index.
        """
        self.set_from_label_and_value_arrays(
            history.parameter_labels, history.get_parameters(index)
        )

    def update_parameter_expression(self):
        """Update all parameters which have an expression.

        Raises
        ------
        ValueError
            Raised if an expression evaluates to a non-numeric value.
        """
        for label, parameter in self.all():
            if parameter.expression is not None:
                value = self._evaluator(parameter.transformed_expression)
                if not isinstance(value, (int, float)):
                    raise ValueError(
                        f"Expression '{parameter.expression}' of parameter '{label}' evaluates to "
                        f"non numeric value '{value}'."
                    )
                parameter.value = value

    def markdown(self) -> MarkdownStr:
        """Format the :class:`ParameterGroup` as markdown string.

        This is done by recursing the nested :class:`ParameterGroup` tree.

        Returns
        -------
        MarkdownStr :
            The markdown representation as string.
        """
        node_indentation = "  " * self.get_nr_roots()
        return_string = ""
        table_header = [
            "_Label_",
            "_Value_",
            "_StdErr_",
            "_Min_",
            "_Max_",
            "_Vary_",
            "_Non-Negative_",
            "_Expr_",
        ]
        if self.label is not None:
            return_string += f"{node_indentation}* __{self.label}__:\n"
        if len(self._parameters):
            parameter_rows = [
                [
                    parameter.label,
                    parameter.value,
                    parameter.standard_error,
                    parameter.minimum,
                    parameter.maximum,
                    parameter.vary,
                    parameter.non_negative,
                    parameter.expression,
                ]
                for _, parameter in self._parameters.items()
            ]
            parameter_table = indent(
                tabulate(
                    parameter_rows, headers=table_header, tablefmt="github", missingval="None"
                ),
                f"  {node_indentation}",
            )
            return_string += f"\n{parameter_table}\n\n"
        for _, child_group in sorted(self.items()):
            return_string += f"{child_group.__str__()}"
        return MarkdownStr(return_string)

    def _repr_markdown_(self) -> str:
        """Create a markdown respresentation.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str :
            The markdown representation as string.
        """
        return str(self.markdown())

    def __repr__(self) -> str:
        """Representation used by repl and tracebacks.

        Returns
        -------
        str :
            A string representation of the :class:`ParameterGroup`.
        """
        parameter_short_notations = [
            [str(parameter.label), parameter.value] for parameter in self._parameters.values()
        ]
        if self.label is None:
            if len(self._parameters) == 0:
                return f"{type(self).__name__}.from_dict({super().__repr__()})"
            else:
                return f"{type(self).__name__}.from_list({parameter_short_notations})"
        if len(self._parameters):
            return parameter_short_notations.__repr__()
        else:
            return super().__repr__()

    def __str__(self) -> str:
        """Representation used by print and str."""
        return str(self.markdown())



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Literal

from glotaran.model.dataset_model import DatasetModel


@dataclass
class DatasetGroupModel:
    """A group of datasets which will evaluated independently."""

    residual_function: Literal[
        "variable_projection", "non_negative_least_squares"
    ] = "variable_projection"
    """The residual function to use."""

    link_clp: bool | None = None
    """Whether to link the clp parameter."""


@dataclass
class DatasetGroup:
    model: DatasetGroupModel
    dataset_models: dict[str, DatasetModel] = field(default_factory=dict)



"""Implementation of the folder Io plugin.

The current implementation is an exact copy of how ``Result.save(path)``
worked in glotaran 0.3.x and meant as an compatibility function.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from glotaran.io import save_dataset
from glotaran.io import save_model
from glotaran.io import save_parameters
from glotaran.io import save_scheme
from glotaran.io.interface import ProjectIoInterface
from glotaran.plugin_system.project_io_registration import SAVING_OPTIONS_DEFAULT
from glotaran.plugin_system.project_io_registration import register_project_io

if TYPE_CHECKING:
    from glotaran.plugin_system.project_io_registration import SavingOptions
    from glotaran.project import Result


@register_project_io(["folder", "legacy"])
class FolderProjectIo(ProjectIoInterface):
    """Project Io plugin to save result data to a folder.

    There won't be a serialization of the Result object, but simply
    a markdown summary output and the important data saved to files.
    """

    def save_result(
        self,
        result: Result,
        result_path: str,
        *,
        saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
    ) -> list[str]:
        """Save the result to a given folder.

        Returns a list with paths of all saved items.
        The following files are saved if not configured otherwise:
        * `result.md`: The result with the model formatted as markdown text.
        * `model.yml`: Model spec file.
        * `scheme.yml`: Scheme spec file.
        * `initial_parameters.csv`: Initially used parameters.
        * `optimized_parameters.csv`: The optimized parameter as csv file.
        * `parameter_history.csv`: Parameter changes over the optimization
        * `{dataset_label}.nc`: The result data for each dataset as NetCDF file.

        Note
        ----
        As a side effect it populates the file path properties of ``result`` which can be
        used in other plugins (e.g. the ``yml`` save_result).

        Parameters
        ----------
        result : Result
            Result instance to be saved.
        result_path : str
            The path to the folder in which to save the result.
        saving_options : SavingOptions
            Options for saving the the result.


        Returns
        -------
        list[str]
            List of file paths which were created.

        Raises
        ------
        ValueError
            If ``result_path`` is a file.
        """
        result_folder = Path(result_path)
        if result_folder.is_file():
            raise ValueError(f"The path '{result_folder}' is not a directory.")
        result_folder.mkdir(parents=True, exist_ok=True)

        paths = []
        if saving_options.report:
            report_file = result_folder / "result.md"
            report_file.write_text(str(result.markdown()))
            paths.append(report_file.as_posix())

        result.scheme.model_file = "model.yml"
        save_model(
            result.scheme.model, result_folder / result.scheme.model_file, allow_overwrite=True
        )
        paths.append((result_folder / result.scheme.model_file).as_posix())

        result.initial_parameters_file = (
            result.scheme.parameters_file
        ) = f"initial_parameters.{saving_options.parameter_format}"
        save_parameters(
            result.scheme.parameters,
            result_folder / result.scheme.parameters_file,
            format_name=saving_options.parameter_format,
            allow_overwrite=True,
        )
        paths.append((result_folder / result.scheme.parameters_file).as_posix())

        result.optimized_parameters_file = (
            f"optimized_parameters.{saving_options.parameter_format}"
        )
        save_parameters(
            result.optimized_parameters,
            result_folder / result.optimized_parameters_file,
            format_name=saving_options.parameter_format,
            allow_overwrite=True,
        )
        paths.append((result_folder / result.optimized_parameters_file).as_posix())

        result.scheme_file = "scheme.yml"
        save_scheme(result.scheme, result_folder / result.scheme_file, allow_overwrite=True)
        paths.append((result_folder / result.scheme_file).as_posix())

        result.parameter_history_file = "parameter_history.csv"
        result.parameter_history.to_csv(result_folder / result.parameter_history_file)
        paths.append((result_folder / result.parameter_history_file).as_posix())

        result.data_files = {
            label: f"{label}.{saving_options.data_format}" for label in result.data
        }

        for label, data_file in result.data_files.items():
            save_dataset(
                result.data[label],
                result_folder / data_file,
                format_name=saving_options.data_format,
                allow_overwrite=True,
            )
            paths.append((result_folder / data_file).as_posix())

        return paths



"""Deprecation helpers and place to put deprecated implementations till removing."""
from glotaran.deprecation.deprecation_utils import deprecate
from glotaran.deprecation.deprecation_utils import deprecate_dict_entry
from glotaran.deprecation.deprecation_utils import deprecate_module_attribute
from glotaran.deprecation.deprecation_utils import deprecate_submodule
from glotaran.deprecation.deprecation_utils import raise_deprecation_error
from glotaran.deprecation.deprecation_utils import warn_deprecated



"""The glotaran parameter package."""
from glotaran.parameter.parameter import Parameter
from glotaran.parameter.parameter_group import ParameterGroup
from glotaran.parameter.parameter_history import ParameterHistory



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from glotaran.io import save_result
from glotaran.project.test.test_result import dummy_result  # noqa: F401

if TYPE_CHECKING:
    from typing import Literal

    from glotaran.project.result import Result


@pytest.mark.parametrize("format_name", ("folder", "legacy"))
def test_save_result_folder(
    tmp_path: Path,
    dummy_result: Result,  # noqa: F811
    format_name: Literal["folder", "legacy"],
):
    """Check all files exist."""

    result_dir = tmp_path / "testresult"
    save_paths = save_result(
        result_path=str(result_dir), format_name=format_name, result=dummy_result
    )

    wanted_files = [
        "result.md",
        "scheme.yml",
        "model.yml",
        "initial_parameters.csv",
        "optimized_parameters.csv",
        "parameter_history.csv",
        "dataset1.nc",
        "dataset2.nc",
        "dataset3.nc",
    ]
    for wanted in wanted_files:
        assert (result_dir / wanted).exists()
        assert (result_dir / wanted).as_posix() in save_paths


@pytest.mark.parametrize("format_name", ("folder", "legacy"))
def test_save_result_folder_error_path_is_file(
    tmp_path: Path,
    dummy_result: Result,  # noqa: F811
    format_name: Literal["folder", "legacy"],
):
    """Raise error if result_path is a file without extension and overwrite is true."""

    result_dir = tmp_path / "testresult"
    result_dir.touch()

    with pytest.raises(ValueError, match="The path '.+?' is not a directory."):
        save_result(
            result_path=str(result_dir),
            format_name=format_name,
            result=dummy_result,
            allow_overwrite=True,
        )



"""Test deprecated functionality in 'glotaran.project.result'."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.project.test.test_result import dummy_result  # noqa: F401

if TYPE_CHECKING:

    from glotaran.project.result import Result


def test_Result_get_dataset_method(dummy_result: Result):  # noqa: F811
    """Result.get_dataset(dataset_label) gives correct dataset."""

    _, result = deprecation_warning_on_call_test_helper(
        dummy_result.get_dataset, args=["dataset1"], raise_exception=True
    )

    assert result == dummy_result.data["dataset1"]


def test_Result_get_dataset_method_error(dummy_result: Result):  # noqa: F811
    """Result.get_dataset(dataset_label) error on wrong key."""

    with pytest.raises(ValueError, match="Unknown dataset 'foo'"):
        deprecation_warning_on_call_test_helper(
            dummy_result.get_dataset, args=["foo"], raise_exception=True
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

if TYPE_CHECKING:
    from glotaran.analysis.optimization_group import OptimizationGroup


class OptimizationGroupCalculator:
    """A Problem class"""

    def __init__(self, group: OptimizationGroup):
        self._group = group

    def calculate_matrices(self):
        raise NotImplementedError

    def calculate_residual(self):
        raise NotImplementedError

    def calculate_full_penalty(self) -> np.ndarray:
        raise NotImplementedError

    def prepare_result_creation(self):
        pass

    def create_index_dependent_result_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        """Creates a result datasets for index dependent matrices."""
        raise NotImplementedError

    def create_index_independent_result_dataset(
        self, label: str, dataset: xr.Dataset
    ) -> xr.Dataset:
        """Creates a result datasets for index independent matrices."""
        raise NotImplementedError



"""Baseclasses to create Data/Project IO plugins from.

The main purpose of those classes are to guarantee a consistent API via
typechecker like ``mypy`` and demonstarate with methods are accessed by
highlevel convenience functions for a given type of plugin.

To add additional options to a method, those options need to be
keyword only arguments.
See: https://www.python.org/dev/peps/pep-3102/

"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable
    from typing import Union

    import xarray as xr

    from glotaran.model import Model
    from glotaran.parameter import ParameterGroup
    from glotaran.project import Result
    from glotaran.project import Scheme

    DataLoader = Callable[[str], Union[xr.Dataset, xr.DataArray]]
    DataSaver = Callable[[str, Union[xr.Dataset, xr.DataArray]], None]


class DataIoInterface:
    """Baseclass for Data IO plugins."""

    def __init__(self, format_name: str) -> None:
        """Initialize a Data IO plugin with the name of the format.

        Parameters
        ----------
        format_name : str
            Name of the supported format an instance uses.
        """
        self.format = format_name

    def load_dataset(self, file_name: str) -> xr.Dataset | xr.DataArray:
        """Read data from a file to :xarraydoc:`Dataset` or :xarraydoc:`DataArray` (**NOT IMPLEMENTED**).

        Parameters
        ----------
        file_name : str
            File containing the data.

        Returns
        -------
        xr.Dataset|xr.DataArray
            Data loaded from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"""Cannot read data with format: {self.format!r}""")

    def save_dataset(
        self,
        dataset: xr.Dataset | xr.DataArray,
        file_name: str,
    ):
        """Save data from :xarraydoc:`Dataset` to a file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        dataset : xr.Dataset
            Dataset to be saved to file.
        file_name : str
            File to write the data to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"""Cannot save data with format: {self.format!r}""")


class ProjectIoInterface:
    """Baseclass for Project IO plugins."""

    def __init__(self, format_name: str) -> None:
        """Initialize a Project IO plugin with the name of the format.

        Parameters
        ----------
        format_name : str
            Name of the supported format an instance uses.
        """
        self.format = format_name

    def load_model(self, file_name: str) -> Model:
        """Create a Model instance from the specs defined in a file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        file_name : str
            File containing the model specs.

        Returns
        -------
        Model
            Model instance created from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read models with format {self.format!r}")

    def save_model(self, model: Model, file_name: str):
        """Save a Model instance to a spec file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        model: Model
            Model instance to save to specs file.
        file_name : str
            File to write the model specs to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save models with format {self.format!r}")

    def load_parameters(self, file_name: str) -> ParameterGroup:
        """Create a ParameterGroup instance from the specs defined in a file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        file_name : str
            File containing the parameter specs.

        Returns
        -------
        ParameterGroup
            ParameterGroup instance created from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read parameters with format {self.format!r}")

    def save_parameters(self, parameters: ParameterGroup, file_name: str):
        """Save a ParameterGroup instance to a spec file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        parameters : ParameterGroup
            ParameterGroup instance to save to specs file.
        file_name : str
            File to write the parameter specs to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save parameters with format {self.format!r}")

    def load_scheme(self, file_name: str) -> Scheme:
        """Create a Scheme instance from the specs defined in a file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        file_name : str
            File containing the parameter specs.

        Returns
        -------
        Scheme
            Scheme instance created from the file.

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read scheme with format {self.format!r}")

    def save_scheme(self, scheme: Scheme, file_name: str):
        """Save a Scheme instance to a spec file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        scheme : Scheme
            Scheme instance to save to specs file.
        file_name : str
            File to write the scheme specs to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save scheme with format {self.format!r}")

    def load_result(self, result_path: str) -> Result:
        """Create a Result instance from the specs defined in a file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        result_path : str
            Path containing the result data.

        Returns
        -------
        Result
            Result instance created from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read result with format {self.format!r}")

    def save_result(self, result: Result, result_path: str) -> list[str] | None:
        """Save a Result instance to a spec file (**NOT IMPLEMENTED**).

        Parameters
        ----------
        result : Result
            Result instance to save to specs file.
        result_path : str
            Path to write the result data to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save result with format {self.format!r}")



"""This package contains the kinetic megacomplex item."""
from __future__ import annotations

import numba as nb
import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.irf import Irf
from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import megacomplex
from glotaran.parameter import Parameter


@megacomplex(
    dimension="time",
    unique=True,
    properties={
        "order": {"type": int},
        "width": {"type": Parameter, "allow_none": True},
    },
    dataset_model_items={
        "irf": {"type": Irf, "allow_none": True},
    },
    register_as="coherent-artifact",
)
class CoherentArtifactMegacomplex(Megacomplex):
    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        indices: dict[str, int],
        **kwargs,
    ):
        if not 1 <= self.order <= 3:
            raise ModelError("Coherent artifact order must be between in [1,3]")

        if dataset_model.irf is None:
            raise ModelError(f'No irf in dataset "{dataset_model.label}"')

        if not isinstance(dataset_model.irf, IrfMultiGaussian):
            raise ModelError(f'Irf in dataset "{dataset_model.label} is not a gaussian irf."')

        global_dimension = dataset_model.get_global_dimension()
        global_index = indices.get(global_dimension)
        global_axis = dataset_model.get_global_axis()
        model_axis = dataset_model.get_model_axis()

        irf = dataset_model.irf

        center, width, _, _, _, _ = irf.parameter(global_index, global_axis)
        center = center[0]
        width = self.width.value if self.width is not None else width[0]

        matrix = _calculate_coherent_artifact_matrix(center, width, model_axis, self.order)
        return self.compartments(), matrix

    def compartments(self):
        return [f"coherent_artifact_{i}" for i in range(1, self.order + 1)]

    def index_dependent(self, dataset_model: DatasetModel) -> bool:
        return False

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        if not is_full_model:
            global_dimension = dataset_model.get_global_dimension()
            model_dimension = dataset_model.get_model_dimension()
            dataset.coords["coherent_artifact_order"] = np.arange(1, self.order + 1)
            response_dimensions = (model_dimension, "coherent_artifact_order")
            if dataset_model.is_index_dependent() is True:
                response_dimensions = (global_dimension, *response_dimensions)
            dataset["coherent_artifact_response"] = (
                response_dimensions,
                dataset.matrix.sel(clp_label=self.compartments()).values,
            )
            dataset["coherent_artifact_associated_spectra"] = (
                (global_dimension, "coherent_artifact_order"),
                dataset.clp.sel(clp_label=self.compartments()).values,
            )


@nb.jit(nopython=True, parallel=True)
def _calculate_coherent_artifact_matrix(center, width, axis, order):
    matrix = np.zeros((axis.size, order), dtype=np.float64)

    matrix[:, 0] = np.exp(-1 * (axis - center) ** 2 / (2 * width ** 2))
    if order > 1:
        matrix[:, 1] = matrix[:, 0] * (center - axis) / width ** 2

    if order > 2:
        matrix[:, 2] = (
            matrix[:, 0] * (center ** 2 - width ** 2 - 2 * center * axis + axis ** 2) / width ** 4
        )
    return matrix



from __future__ import annotations

from typing import List

import numpy as np

from glotaran.model import Megacomplex
from glotaran.model import Model
from glotaran.model import megacomplex
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup


@megacomplex(dimension="global", properties={})
class SimpleTestMegacomplexGlobal(Megacomplex):
    def calculate_matrix(self, dataset_model, indices, **kwargs):
        axis = dataset_model.get_coordinates()
        assert "model" in axis
        assert "global" in axis
        axis = axis["global"]
        compartments = ["s1", "s2"]
        r_compartments = []
        array = np.zeros((axis.shape[0], len(compartments)))

        for i in range(len(compartments)):
            r_compartments.append(compartments[i])
            for j in range(axis.shape[0]):
                array[j, i] = (i + j) * axis[j]
        return r_compartments, array

    def index_dependent(self, dataset_model):
        return False


@megacomplex(dimension="model", properties={"is_index_dependent": bool})
class SimpleTestMegacomplex(Megacomplex):
    def calculate_matrix(self, dataset_model, indices, **kwargs):
        axis = dataset_model.get_coordinates()
        assert "model" in axis
        assert "global" in axis

        axis = axis["model"]
        compartments = ["s1", "s2"]
        r_compartments = []
        array = np.zeros((axis.shape[0], len(compartments)))

        for i in range(len(compartments)):
            r_compartments.append(compartments[i])
            for j in range(axis.shape[0]):
                array[j, i] = (i + j) * axis[j]
        return r_compartments, array

    def index_dependent(self, dataset_model):
        return self.is_index_dependent


class SimpleTestModel(Model):
    @classmethod
    def from_dict(
        cls,
        model_dict,
        *,
        megacomplex_types: dict[str, type[Megacomplex]] | None = None,
        default_megacomplex_type: str | None = None,
    ):
        defaults: dict[str, type[Megacomplex]] = {
            "model_complex": SimpleTestMegacomplex,
            "global_complex": SimpleTestMegacomplexGlobal,
        }
        if megacomplex_types is not None:
            defaults.update(megacomplex_types)
        return super().from_dict(
            model_dict,
            megacomplex_types=defaults,
            default_megacomplex_type=default_megacomplex_type,
        )


@megacomplex(
    dimension="model",
    properties={"is_index_dependent": bool},
    dataset_properties={
        "kinetic": List[Parameter],
    },
)
class SimpleKineticMegacomplex(Megacomplex):
    def calculate_matrix(self, dataset_model, indices, **kwargs):
        axis = dataset_model.get_coordinates()
        axis = axis["model"]
        kinpar = -1 * np.asarray(dataset_model.kinetic)
        if dataset_model.label == "dataset3":
            # this case is for the ThreeDatasetDecay test
            compartments = [f"s{i+2}" for i in range(len(kinpar))]
        else:
            compartments = [f"s{i+1}" for i in range(len(kinpar))]
        array = np.exp(np.outer(axis, kinpar))
        return compartments, array

    def index_dependent(self, dataset_model):
        return self.is_index_dependent

    def finalize_data(
        self,
        dataset_model,
        dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        pass


@megacomplex(dimension="global", properties={})
class SimpleSpectralMegacomplex(Megacomplex):
    def calculate_matrix(self, dataset_model, indices, **kwargs):
        axis = dataset_model.get_coordinates()
        axis = axis["global"]
        kinpar = dataset_model.kinetic
        if dataset_model.label == "dataset3":
            # this case is for the ThreeDatasetDecay test
            compartments = [f"s{i+2}" for i in range(len(kinpar))]
        else:
            compartments = [f"s{i+1}" for i in range(len(kinpar))]
        array = np.asarray([[1 for _ in range(axis.size)] for _ in compartments]).T
        return compartments, array

    def index_dependent(self, dataset_model):
        return False


@megacomplex(
    dimension="global",
    properties={
        "location": {"type": List[Parameter], "allow_none": True},
        "amplitude": {"type": List[Parameter], "allow_none": True},
        "delta": {"type": List[Parameter], "allow_none": True},
    },
)
class ShapedSpectralMegacomplex(Megacomplex):
    def calculate_matrix(self, dataset_model, indices, **kwargs):
        location = np.asarray(self.location)
        amp = np.asarray(self.amplitude)
        delta = np.asarray(self.delta)

        axis = dataset_model.get_coordinates()
        axis = axis["global"]
        array = np.empty((location.size, axis.size), dtype=np.float64)

        for i in range(location.size):
            array[i, :] = amp[i] * np.exp(
                -np.log(2) * np.square(2 * (axis - location[i]) / delta[i])
            )
        compartments = [f"s{i+1}" for i in range(location.size)]
        return compartments, array.T

    def index_dependent(self, dataset_model):
        return False

    def finalize_data(
        self,
        dataset_model,
        dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        pass


class DecayModel(Model):
    @classmethod
    def from_dict(
        cls,
        model_dict,
        *,
        megacomplex_types: dict[str, type[Megacomplex]] | None = None,
        default_megacomplex_type: str | None = None,
    ):
        defaults: dict[str, type[Megacomplex]] = {
            "model_complex": SimpleKineticMegacomplex,
            "global_complex": SimpleSpectralMegacomplex,
            "global_complex_shaped": ShapedSpectralMegacomplex,
        }
        if megacomplex_types is not None:
            defaults.update(megacomplex_types)
        return super().from_dict(
            model_dict,
            megacomplex_types=defaults,
            default_megacomplex_type=default_megacomplex_type,
        )


class OneCompartmentDecay:
    scale = 2
    wanted_parameters = ParameterGroup.from_list([101e-4])
    initial_parameters = ParameterGroup.from_list([100e-5, [scale, {"vary": False}]])

    global_axis = np.asarray([1.0])
    model_axis = np.arange(0, 150, 1.5)

    sim_model_dict = {
        "megacomplex": {"m1": {"is_index_dependent": False}, "m2": {"type": "global_complex"}},
        "dataset": {
            "dataset1": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1"],
            }
        },
    }
    sim_model = DecayModel.from_dict(sim_model_dict)
    model_dict = {
        "megacomplex": {"m1": {"is_index_dependent": False}},
        "dataset": {
            "dataset1": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "kinetic": ["1"],
                "scale": "2",
            }
        },
    }
    model_dict["dataset"]["dataset1"]["scale"] = "2"
    model = DecayModel.from_dict(model_dict)


class TwoCompartmentDecay:
    wanted_parameters = ParameterGroup.from_list([11e-4, 22e-5])
    initial_parameters = ParameterGroup.from_list([10e-4, 20e-5])

    global_axis = np.asarray([1.0])
    model_axis = np.arange(0, 150, 1.5)

    sim_model = DecayModel.from_dict(
        {
            "megacomplex": {"m1": {"is_index_dependent": False}, "m2": {"type": "global_complex"}},
            "dataset": {
                "dataset1": {
                    "initial_concentration": [],
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "kinetic": ["1", "2"],
                }
            },
        }
    )
    model = DecayModel.from_dict(
        {
            "megacomplex": {"m1": {"is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "initial_concentration": [],
                    "megacomplex": ["m1"],
                    "kinetic": ["1", "2"],
                }
            },
        }
    )


class ThreeDatasetDecay:
    wanted_parameters = ParameterGroup.from_list([101e-4, 201e-3])
    initial_parameters = ParameterGroup.from_list([100e-5, 200e-3])

    global_axis = np.asarray([1.0])
    model_axis = np.arange(0, 150, 1.5)

    global_axis2 = np.asarray([1.0, 2.01])
    model_axis2 = np.arange(0, 100, 1.5)

    global_axis3 = np.asarray([0.99, 3.0])
    model_axis3 = np.arange(0, 150, 1.5)

    sim_model_dict = {
        "megacomplex": {"m1": {"is_index_dependent": False}, "m2": {"type": "global_complex"}},
        "dataset": {
            "dataset1": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1"],
            },
            "dataset2": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1", "2"],
            },
            "dataset3": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["2"],
            },
        },
    }
    sim_model = DecayModel.from_dict(sim_model_dict)

    model_dict = {
        "megacomplex": {"m1": {"is_index_dependent": False}},
        "dataset": {
            "dataset1": {"initial_concentration": [], "megacomplex": ["m1"], "kinetic": ["1"]},
            "dataset2": {
                "initial_concentration": [],
                "megacomplex": ["m1"],
                "kinetic": ["1", "2"],
            },
            "dataset3": {"initial_concentration": [], "megacomplex": ["m1"], "kinetic": ["2"]},
        },
    }
    model = DecayModel.from_dict(model_dict)


class MultichannelMulticomponentDecay:
    wanted_parameters = ParameterGroup.from_dict(
        {
            "k": [0.006, 0.003, 0.0003, 0.03],
            "loc": [
                ["1", 14705],
                ["2", 13513],
                ["3", 14492],
                ["4", 14388],
            ],
            "amp": [
                ["1", 1],
                ["2", 2],
                ["3", 5],
                ["4", 20],
            ],
            "del": [
                ["1", 400],
                ["2", 100],
                ["3", 300],
                ["4", 200],
            ],
        }
    )
    initial_parameters = ParameterGroup.from_dict({"k": [0.006, 0.003, 0.0003, 0.03]})

    global_axis = np.arange(12820, 15120, 50)
    model_axis = np.arange(0, 150, 1.5)

    sim_model = DecayModel.from_dict(
        {
            "megacomplex": {
                "m1": {"is_index_dependent": False},
                "m2": {
                    "type": "global_complex_shaped",
                    "location": ["loc.1", "loc.2", "loc.3", "loc.4"],
                    "delta": ["del.1", "del.2", "del.3", "del.4"],
                    "amplitude": ["amp.1", "amp.2", "amp.3", "amp.4"],
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )
    model = DecayModel.from_dict(
        {
            "megacomplex": {"m1": {"is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )


class FullModel:
    model = DecayModel.from_dict(
        {
            "megacomplex": {
                "m1": {"is_index_dependent": False},
                "m2": {
                    "type": "global_complex_shaped",
                    "location": ["loc.1", "loc.2", "loc.3", "loc.4"],
                    "delta": ["del.1", "del.2", "del.3", "del.4"],
                    "amplitude": ["amp.1", "amp.2", "amp.3", "amp.4"],
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )
    parameters = ParameterGroup.from_dict(
        {
            "k": [0.006, 0.003, 0.0003, 0.03],
            "loc": [
                ["1", 14705],
                ["2", 13513],
                ["3", 14492],
                ["4", 14388],
            ],
            "amp": [
                ["1", 1],
                ["2", 2],
                ["3", 5],
                ["4", 20],
            ],
            "del": [
                ["1", 400],
                ["2", 100],
                ["3", 300],
                ["4", 200],
            ],
        }
    )
    global_axis = np.arange(12820, 15120, 50)
    model_axis = np.arange(0, 150, 1.5)
    coordinates = {"global": global_axis, "model": model_axis}



from dataclasses import dataclass

from glotaran.project.dataclass_helpers import asdict
from glotaran.project.dataclass_helpers import exclude_from_dict_field
from glotaran.project.dataclass_helpers import file_representation_field
from glotaran.project.dataclass_helpers import fromdict


def dummy_loader(file: str) -> int:
    return {"foo.file": 21, "bar.file": 42}[file]


def test_serialize_to_file_name_field():
    @dataclass
    class DummyDataclass:
        foo: int = exclude_from_dict_field()
        foo_file: int = file_representation_field("foo", dummy_loader)
        bar: int = exclude_from_dict_field(default=42)
        bar_file: int = file_representation_field("bar", dummy_loader, default="bar.file")
        baz: int = 84

    dummy_class = DummyDataclass(foo=21, foo_file="foo.file")

    dummy_class_dict = asdict(dummy_class)

    assert "foo" not in dummy_class_dict
    assert dummy_class_dict["foo_file"] == "foo.file"

    assert "bar" not in dummy_class_dict
    assert dummy_class_dict["bar_file"] == "bar.file"

    assert dummy_class_dict["baz"] == 84
    assert dummy_class_dict["baz"] == dummy_class.baz

    loaded = fromdict(DummyDataclass, dummy_class_dict)

    assert loaded == dummy_class



"""The model property class."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from glotaran.model.util import wrap_func_as_method
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup

if TYPE_CHECKING:
    from glotaran.model.model import Model


class ModelProperty(property):
    def __init__(self, cls, name, prop_type, doc, default, allow_none):

        self._name = name
        self._allow_none = allow_none
        self._determine_if_parameter(prop_type)

        self._type = prop_type if not self._is_parameter else Union[str, prop_type]

        @wrap_func_as_method(cls, name=name)
        def setter(that_self, value: self._type):
            if value is None and not self._allow_none:
                raise Exception(
                    f"Property '{name}' of '{cls.__name__}' is not allowed to set to None."
                )
            if self._is_parameter and value is not None:
                if self._is_parameter_value and isinstance(value, str):
                    value = Parameter(full_label=str(value))
                elif self._is_parameter_list and all(isinstance(v, str) for v in value):
                    value = [Parameter(full_label=str(v)) for v in value]
                elif self._is_parameter_dict and all(isinstance(v, str) for v in value.values()):
                    for k, v in value.items():
                        value[k] = Parameter(full_label=v)
            setattr(that_self, f"_{self._name}", value)

        @wrap_func_as_method(cls, name=name)
        def getter(that_self) -> prop_type:
            value = getattr(that_self, f"_{self._name}")
            if value is None:
                value = default
            return value

        super().__init__(fget=getter, fset=setter, doc=doc)

    @property
    def allow_none(self) -> bool:
        return self._allow_none

    @property
    def property_type(self) -> type:
        return self._type

    def as_dict_value(self, value):
        if value is None:
            return None
        elif self._is_parameter_value:
            return value.full_label
        elif self._is_parameter_list:
            return [v.full_label for v in value]
        elif self._is_parameter_dict:
            return {k: v.full_label for k, v in value.items()}
        return value

    def validate(self, value: Any, model: Model, parameters: ParameterGroup = None) -> list[str]:

        if value is None and self.allow_none:
            return []

        missing_model = []
        if self._name in model.model_items:
            items = getattr(model, self._name)

            if isinstance(value, list):
                for item in value:
                    if item not in items:
                        missing_model.append((self._name, item))
            elif isinstance(value, dict):
                for item in value.values():
                    if item not in items:
                        missing_model.append((self._name, item))
            elif value not in items:
                missing_model.append((self._name, value))
        missing_model = [
            f"Missing Model Item: '{name}'['{label}']" for name, label in missing_model
        ]

        missing_parameters = []
        if parameters is not None and self._is_parameter and value is not None:
            if self._is_parameter_value:
                if not parameters.has(value.full_label):
                    missing_parameters.append(value.full_label)
            elif self._is_parameter_list:
                for item in value:
                    if not parameters.has(item.full_label):
                        missing_parameters.append(item.full_label)
            elif self._is_parameter_dict:
                for item in value.values():
                    if not parameters.has(item.full_label):
                        missing_parameters.append(item.full_label)
        missing_parameters = [f"Missing Parameter: '{p}'" for p in missing_parameters]

        return missing_model + missing_parameters

    def fill(self, value: Any, model: Model, parameter: ParameterGroup) -> Any:

        if value is None:
            return None

        if self._is_parameter:

            if self._is_parameter_value:
                value.set_from_group(parameter)

            elif self._is_parameter_list:
                for v in value:
                    v.set_from_group(parameter)

            elif self._is_parameter_dict:
                for v in value.values():
                    v.set_from_group(parameter)

        elif hasattr(model, self._name):
            if isinstance(value, list):
                value = [getattr(model, self._name)[v].fill(model, parameter) for v in value]
            elif isinstance(value, dict):
                value = {
                    k: getattr(model, self._name)[v].fill(model, parameter)
                    for (k, v) in value.items()
                }
            elif not isinstance(value, bool):
                value = getattr(model, self._name)[value].fill(model, parameter)

        return value

    def get_parameters(self, value: Any) -> list[str]:
        if value is None:
            return []
        elif self._is_parameter_value:
            return [value.full_label]
        elif self._is_parameter_list:
            return [v.full_label for v in value]
        elif self._is_parameter_dict:
            return [v.full_label for v in value.values()]
        return []

    def _determine_if_parameter(self, type):
        self._is_parameter_value = type is Parameter
        self._is_parameter_list = (
            hasattr(type, "__origin__")
            and issubclass(type.__origin__, List)
            and type.__args__[0] is Parameter
        )
        self._is_parameter_dict = (
            hasattr(type, "__origin__")
            and issubclass(type.__origin__, Dict)
            and type.__args__[1] is Parameter
        )
        self._is_parameter = (
            self._is_parameter_value or self._is_parameter_list or self._is_parameter_dict
        )



"""Functions for data IO

Note:
-----
Since Io functionality is purely plugin based this package mostly
reexports functions from the pluginsystem from a common place.
"""

from glotaran.io.interface import DataIoInterface
from glotaran.io.interface import ProjectIoInterface
from glotaran.io.prepare_dataset import prepare_time_trace_dataset
from glotaran.plugin_system.data_io_registration import data_io_plugin_table
from glotaran.plugin_system.data_io_registration import get_dataloader
from glotaran.plugin_system.data_io_registration import get_datasaver
from glotaran.plugin_system.data_io_registration import load_dataset
from glotaran.plugin_system.data_io_registration import register_data_io
from glotaran.plugin_system.data_io_registration import save_dataset
from glotaran.plugin_system.data_io_registration import set_data_plugin
from glotaran.plugin_system.data_io_registration import show_data_io_method_help
from glotaran.plugin_system.project_io_registration import SAVING_OPTIONS_DEFAULT
from glotaran.plugin_system.project_io_registration import SAVING_OPTIONS_MINIMAL
from glotaran.plugin_system.project_io_registration import SavingOptions
from glotaran.plugin_system.project_io_registration import get_project_io_method
from glotaran.plugin_system.project_io_registration import load_model
from glotaran.plugin_system.project_io_registration import load_parameters
from glotaran.plugin_system.project_io_registration import load_result
from glotaran.plugin_system.project_io_registration import load_scheme
from glotaran.plugin_system.project_io_registration import project_io_plugin_table
from glotaran.plugin_system.project_io_registration import register_project_io
from glotaran.plugin_system.project_io_registration import save_model
from glotaran.plugin_system.project_io_registration import save_parameters
from glotaran.plugin_system.project_io_registration import save_result
from glotaran.plugin_system.project_io_registration import save_scheme
from glotaran.plugin_system.project_io_registration import set_project_plugin
from glotaran.plugin_system.project_io_registration import show_project_io_method_help


def __getattr__(attribute_name: str):
    from glotaran.deprecation.deprecation_utils import deprecate_module_attribute

    if attribute_name == "read_data_file":
        return deprecate_module_attribute(
            deprecated_qual_name="glotaran.io.read_data_file",
            new_qual_name="glotaran.io.load_dataset",
            to_be_removed_in_version="0.6.0",
        )

    raise AttributeError(f"module {__name__} has no attribute {attribute_name}")



import numpy as np

from glotaran.analysis.simulation import simulate
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme

sim_model = Model.from_dict(
    {
        "initial_concentration": {
            "j1": {
                "compartments": ["s1", "s2", "s3"],
                "parameters": ["j.1", "j.0", "j.0"],
            },
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s2", "s1"): "kinetic.1",
                    ("s3", "s2"): "kinetic.2",
                    ("s3", "s3"): "kinetic.3",
                }
            }
        },
        "megacomplex": {
            "m1": {
                "type": "decay",
                "k_matrix": ["k1"],
            },
            "m2": {
                "type": "spectral",
                "shape": {
                    "s1": "sh1",
                    "s2": "sh2",
                    "s3": "sh3",
                },
            },
        },
        "shape": {
            "sh1": {
                "type": "gaussian",
                "amplitude": "shapes.amps.1",
                "location": "shapes.locs.1",
                "width": "shapes.width.1",
            },
            "sh2": {
                "type": "gaussian",
                "amplitude": "shapes.amps.2",
                "location": "shapes.locs.2",
                "width": "shapes.width.2",
            },
            "sh3": {
                "type": "gaussian",
                "amplitude": "shapes.amps.3",
                "location": "shapes.locs.3",
                "width": "shapes.width.3",
            },
        },
        "irf": {
            "irf1": {"type": "gaussian", "center": "irf.center", "width": "irf.width"},
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "irf": "irf1",
            }
        },
    },
    megacomplex_types={"decay": DecayMegacomplex, "spectral": SpectralMegacomplex},
)

wanted_parameter = ParameterGroup.from_dict(
    {
        "j": [
            ["1", 1, {"non-negative": False, "vary": False}],
            ["0", 0, {"non-negative": False, "vary": False}],
        ],
        "kinetic": [
            ["1", 0.5],
            ["2", 0.3],
            ["3", 0.1],
        ],
        "shapes": {"amps": [30, 20, 40], "locs": [620, 630, 650], "width": [40, 20, 60]},
        "irf": [["center", 0.3], ["width", 0.1]],
    }
)

parameter = ParameterGroup.from_dict(
    {
        "j": [
            ["1", 1, {"vary": False, "non-negative": False}],
            ["0", 0, {"vary": False, "non-negative": False}],
        ],
        "kinetic": [
            ["1", 0.5],
            ["2", 0.3],
            ["3", 0.1],
        ],
        "irf": [["center", 0.3], ["width", 0.1]],
    }
)

_time = np.arange(-1, 20, 0.01)
_spectral = np.arange(600, 700, 1.4)

dataset = simulate(
    sim_model,
    "dataset1",
    wanted_parameter,
    {"time": _time, "spectral": _spectral},
    noise=True,
    noise_std_dev=1e-2,
)

model = Model.from_dict(
    {
        "initial_concentration": {
            "j1": {"compartments": ["s1", "s2", "s3"], "parameters": ["j.1", "j.0", "j.0"]},
        },
        "k_matrix": {
            "k1": {
                "matrix": {
                    ("s2", "s1"): "kinetic.1",
                    ("s3", "s2"): "kinetic.2",
                    ("s3", "s3"): "kinetic.3",
                }
            }
        },
        "megacomplex": {
            "m1": {
                "type": "decay",
                "k_matrix": ["k1"],
            }
        },
        "irf": {
            "irf1": {"type": "gaussian", "center": "irf.center", "width": "irf.width"},
        },
        "dataset": {
            "dataset1": {
                "initial_concentration": "j1",
                "megacomplex": ["m1"],
                "irf": "irf1",
            }
        },
    },
    megacomplex_types={"decay": DecayMegacomplex},
)
scheme = Scheme(model=model, parameters=parameter, data={"dataset1": dataset})



import numpy as np
import xarray as xr

from glotaran.analysis.optimization_group import OptimizationGroup
from glotaran.analysis.test.models import SimpleTestModel
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


def test_single_dataset():
    model = SimpleTestModel.from_dict(
        {
            "megacomplex": {"m1": {"is_index_dependent": False}},
            "dataset_groups": {"default": {"link_clp": True}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )
    print(model.validate())
    assert model.valid()

    parameters = ParameterGroup.from_list([1, 10])
    print(model.validate(parameters))
    assert model.valid(parameters)
    global_axis = [1, 2, 3]
    model_axis = [5, 7, 9, 12]

    data = {
        "dataset1": xr.DataArray(
            np.ones((3, 4)), coords=[("global", global_axis), ("model", model_axis)]
        ).to_dataset(name="data")
    }

    scheme = Scheme(model, parameters, data)
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])
    bag = optimization_group._calculator.bag
    datasets = optimization_group._calculator.groups
    assert len(datasets) == 1
    assert len(bag) == 3
    assert all(p.data.size == 4 for p in bag)
    assert all(p.dataset_models[0].label == "dataset1" for p in bag)
    assert all(all(p.dataset_models[0].axis["model"] == model_axis) for p in bag)
    assert all(all(p.dataset_models[0].axis["global"] == global_axis) for p in bag)
    assert [p.dataset_models[0].indices["global"] for p in bag] == [0, 1, 2]


def test_multi_dataset_no_overlap():
    model = SimpleTestModel.from_dict(
        {
            "megacomplex": {"m1": {"is_index_dependent": False}},
            "dataset_groups": {"default": {"link_clp": True}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
                "dataset2": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )

    model.grouped = lambda: True
    print(model.validate())
    assert model.valid()
    assert model.grouped()

    parameters = ParameterGroup.from_list([1, 10])
    print(model.validate(parameters))
    assert model.valid(parameters)

    global_axis_1 = [1, 2, 3]
    model_axis_1 = [5, 7]
    global_axis_2 = [4, 5, 6]
    model_axis_2 = [5, 7, 9]
    data = {
        "dataset1": xr.DataArray(
            np.ones((3, 2)), coords=[("global", global_axis_1), ("model", model_axis_1)]
        ).to_dataset(name="data"),
        "dataset2": xr.DataArray(
            np.ones((3, 3)), coords=[("global", global_axis_2), ("model", model_axis_2)]
        ).to_dataset(name="data"),
    }

    scheme = Scheme(model, parameters, data)
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])
    bag = list(optimization_group._calculator.bag)
    assert len(optimization_group._calculator.groups) == 2
    assert len(bag) == 6
    assert all(p.data.size == 2 for p in bag[:3])
    assert all(p.dataset_models[0].label == "dataset1" for p in bag[:3])
    assert all(all(p.dataset_models[0].axis["model"] == model_axis_1) for p in bag[:3])
    assert all(all(p.dataset_models[0].axis["global"] == global_axis_1) for p in bag[:3])
    assert [p.dataset_models[0].indices["global"] for p in bag[:3]] == [0, 1, 2]

    assert all(p.data.size == 3 for p in bag[3:])
    assert all(p.dataset_models[0].label == "dataset2" for p in bag[3:])
    assert all(all(p.dataset_models[0].axis["model"] == model_axis_2) for p in bag[3:])
    assert all(all(p.dataset_models[0].axis["global"] == global_axis_2) for p in bag[3:])
    assert [p.dataset_models[0].indices["global"] for p in bag[3:]] == [0, 1, 2]


def test_multi_dataset_overlap():
    model = SimpleTestModel.from_dict(
        {
            "megacomplex": {"m1": {"is_index_dependent": False}},
            "dataset_groups": {"default": {"link_clp": True}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
                "dataset2": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )

    model.grouped = lambda: True
    print(model.validate())
    assert model.valid()
    assert model.grouped()

    parameters = ParameterGroup.from_list([1, 10])
    print(model.validate(parameters))
    assert model.valid(parameters)

    global_axis_1 = [1, 2, 3, 5]
    model_axis_1 = [5, 7]
    global_axis_2 = [0, 1.4, 2.4, 3.4, 9]
    model_axis_2 = [5, 7, 9, 12]
    data = {
        "dataset1": xr.DataArray(
            np.ones((4, 2)), coords=[("global", global_axis_1), ("model", model_axis_1)]
        ).to_dataset(name="data"),
        "dataset2": xr.DataArray(
            np.ones((5, 4)), coords=[("global", global_axis_2), ("model", model_axis_2)]
        ).to_dataset(name="data"),
    }

    scheme = Scheme(model, parameters, data, clp_link_tolerance=5e-1)
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])
    bag = list(optimization_group._calculator.bag)
    assert len(optimization_group._calculator.groups) == 3
    assert "dataset1dataset2" in optimization_group._calculator.groups
    assert optimization_group._calculator.groups["dataset1dataset2"] == ["dataset1", "dataset2"]
    assert len(bag) == 6

    assert all(p.data.size == 4 for p in bag[:1])
    assert all(p.dataset_models[0].label == "dataset1" for p in bag[1:5])
    assert all(all(p.dataset_models[0].axis["model"] == model_axis_1) for p in bag[1:5])
    assert all(all(p.dataset_models[0].axis["global"] == global_axis_1) for p in bag[1:5])
    assert [p.dataset_models[0].indices["global"] for p in bag[1:5]] == [0, 1, 2, 3]

    assert all(p.data.size == 6 for p in bag[1:4])
    assert all(p.dataset_models[1].label == "dataset2" for p in bag[1:4])
    assert all(all(p.dataset_models[1].axis["model"] == model_axis_2) for p in bag[1:4])
    assert all(all(p.dataset_models[1].axis["global"] == global_axis_2) for p in bag[1:4])
    assert [p.dataset_models[1].indices["global"] for p in bag[1:4]] == [1, 2, 3]

    assert all(p.data.size == 4 for p in bag[5:])
    assert bag[4].dataset_models[0].label == "dataset1"
    assert bag[5].dataset_models[0].label == "dataset2"
    assert np.array_equal(bag[4].dataset_models[0].axis["model"], model_axis_1)
    assert np.array_equal(bag[5].dataset_models[0].axis["model"], model_axis_2)
    assert [p.dataset_models[0].indices["global"] for p in bag[1:4]] == [0, 1, 2]



from __future__ import annotations

from pathlib import Path

import yaml

from glotaran.deprecation.modules.builtin_io_yml import model_spec_deprecations
from glotaran.deprecation.modules.builtin_io_yml import scheme_spec_deprecations
from glotaran.io import ProjectIoInterface
from glotaran.io import register_project_io
from glotaran.io import save_result
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project import Result
from glotaran.project import Scheme
from glotaran.project.dataclass_helpers import asdict
from glotaran.project.dataclass_helpers import fromdict
from glotaran.utils.sanitize import sanitize_yaml


@register_project_io(["yml", "yaml", "yml_str"])
class YmlProjectIo(ProjectIoInterface):
    def load_model(self, file_name: str) -> Model:
        """parse_yaml_file reads the given file and parses its content as YML.

        Parameters
        ----------
        filename : str
            filename is the of the file to parse.

        Returns
        -------
        Model
            The content of the file as dictionary.
        """

        spec = self._load_yml(file_name)

        model_spec_deprecations(spec)

        spec = sanitize_yaml(spec)

        default_megacomplex = spec.get("default_megacomplex")

        if default_megacomplex is None and any(
            "type" not in m for m in spec["megacomplex"].values()
        ):
            raise ValueError(
                "Default megacomplex is not defined in model and "
                "at least one megacomplex does not have a type."
            )

        if "megacomplex" not in spec:
            raise ValueError("No megacomplex defined in model")

        return Model.from_dict(spec, megacomplex_types=None, default_megacomplex_type=None)

    def save_model(self, model: Model, file_name: str):
        """Save a Model instance to a spec file.
        Parameters
        ----------
        model: Model
            Model instance to save to specs file.
        file_name : str
            File to write the model specs to.
        """
        model_dict = model.as_dict()
        # We replace tuples with strings
        for items in model_dict.values():
            if not isinstance(items, (list, dict)):
                continue
            item_iterator = items if isinstance(items, list) else items.values()
            for item in item_iterator:
                for prop_name, prop in item.items():
                    if isinstance(prop, dict) and any(isinstance(k, tuple) for k in prop):
                        keys = [f"({k[0]}, {k[1]})" for k in prop]
                        item[prop_name] = {f"{k}": v for k, v in zip(keys, prop.values())}
        _write_dict(file_name, model_dict)

    def load_parameters(self, file_name: str) -> ParameterGroup:
        """Create a ParameterGroup instance from the specs defined in a file.
        Parameters
        ----------
        file_name : str
            File containing the parameter specs.
        Returns
        -------
        ParameterGroup
            ParameterGroup instance created from the file.
        """

        spec = self._load_yml(file_name)

        if isinstance(spec, list):
            return ParameterGroup.from_list(spec)
        else:
            return ParameterGroup.from_dict(spec)

    def load_scheme(self, file_name: str) -> Scheme:
        spec = self._load_yml(file_name)
        scheme_spec_deprecations(spec)
        file_path = Path(file_name)
        return fromdict(Scheme, spec, folder=file_path.parent)

    def save_scheme(self, scheme: Scheme, file_name: str):
        scheme_dict = asdict(scheme)
        _write_dict(file_name, scheme_dict)

    def load_result(self, result_path: str) -> Result:
        """Create a :class:`Result` instance from the specs defined in a file.

        Parameters
        ----------
        result_path : str | PathLike[str]
            Path containing the result data.

        Returns
        -------
        Result
            :class:`Result` instance created from the saved format.
        """
        spec = self._load_yml(result_path)
        return fromdict(Result, spec)

    def save_result(self, result: Result, result_path: str):
        """Write a :class:`Result` instance to a spec file.

        Parameters
        ----------
        result : Result
            :class:`Result` instance to write.
        result_path : str | PathLike[str]
            Path to write the result data to.
        """
        save_result(result, Path(result_path).parent.as_posix(), format_name="folder")
        result_dict = asdict(result)
        _write_dict(result_path, result_dict)

    def _load_yml(self, file_name: str) -> dict:
        if self.format == "yml_str":
            spec = yaml.safe_load(file_name)
        else:
            with open(file_name) as f:
                spec = yaml.safe_load(f)
        return spec


def _write_dict(file_name: str, d: dict):
    with open(file_name, "w") as f:
        f.write(yaml.dump(d))



"""Tests for glotaran.model.dataset_model.DatasetModel"""
from __future__ import annotations

import pytest

from glotaran.builtin.megacomplexes.baseline import BaselineMegacomplex
from glotaran.builtin.megacomplexes.coherent_artifact import CoherentArtifactMegacomplex
from glotaran.builtin.megacomplexes.damped_oscillation import DampedOscillationMegacomplex
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.model.dataset_model import create_dataset_model_type
from glotaran.model.model import default_dataset_properties


class MockModel:
    """Test Model only containing the megacomplex property.

    Multiple and different kinds of megacomplexes are defined
    but only a subset will be used by the DatsetModel.
    """

    def __init__(self) -> None:
        self.megacomplex = {
            # not unique
            "d1": DecayMegacomplex(),
            "d2": DecayMegacomplex(),
            "d3": DecayMegacomplex(),
            "s1": SpectralMegacomplex(),
            "s2": SpectralMegacomplex(),
            "s3": SpectralMegacomplex(),
            "doa1": DampedOscillationMegacomplex(),
            "doa2": DampedOscillationMegacomplex(),
            # unique
            "b1": BaselineMegacomplex(),
            "b2": BaselineMegacomplex(),
            "c1": CoherentArtifactMegacomplex(),
            "c2": CoherentArtifactMegacomplex(),
        }


@pytest.mark.parametrize(
    "used_megacomplexes, expected_problems",
    (
        (
            ["d1"],
            [],
        ),
        (
            ["d1", "d2", "d3"],
            [],
        ),
        (
            ["s1", "s2", "s3"],
            [],
        ),
        (
            ["d1", "d2", "d3", "s1", "s2", "s3", "doa1", "doa2", "b1", "c1"],
            [],
        ),
        (
            ["d1", "b1", "b2"],
            ["Multiple instances of unique megacomplex type 'baseline' in dataset 'ds1'"],
        ),
        (
            ["d1", "c1", "c2"],
            ["Multiple instances of unique megacomplex type 'coherent-artifact' in dataset 'ds1'"],
        ),
        (
            ["d1", "b1", "b2", "c1", "c2"],
            [
                "Multiple instances of unique megacomplex type 'baseline' in dataset 'ds1'",
                "Multiple instances of unique megacomplex type "
                "'coherent-artifact' in dataset 'ds1'",
            ],
        ),
    ),
)
def test_datasetmodel_ensure_unique_megacomplexes(
    used_megacomplexes: list[str], expected_problems: list[str]
):
    """Only report problems if multiple unique megacomplexes of the same type are used."""
    dataset_model = create_dataset_model_type({**default_dataset_properties})()
    dataset_model.megacomplex = used_megacomplexes  # type:ignore
    dataset_model.label = "ds1"  # type:ignore
    problems = dataset_model.ensure_unique_megacomplexes(MockModel())  # type:ignore

    assert len(problems) == len(expected_problems)
    assert problems == expected_problems



"""A base class for global analysis models."""
from __future__ import annotations

import copy
from dataclasses import asdict
from typing import Any
from typing import List
from warnings import warn

import xarray as xr

from glotaran.deprecation import raise_deprecation_error
from glotaran.model.clp_penalties import EqualAreaPenalty
from glotaran.model.constraint import Constraint
from glotaran.model.dataset_group import DatasetGroup
from glotaran.model.dataset_group import DatasetGroupModel
from glotaran.model.dataset_model import create_dataset_model_type
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import create_model_megacomplex_type
from glotaran.model.relation import Relation
from glotaran.model.util import ModelError
from glotaran.model.weight import Weight
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup
from glotaran.plugin_system.megacomplex_registration import get_megacomplex
from glotaran.utils.ipython import MarkdownStr

default_model_items = {
    "clp_area_penalties": EqualAreaPenalty,
    "clp_constraints": Constraint,
    "clp_relations": Relation,
    "weights": Weight,
}

default_dataset_properties = {
    "group": {"type": str, "default": "default"},
    "megacomplex": List[str],
    "megacomplex_scale": {"type": List[Parameter], "allow_none": True},
    "global_megacomplex": {"type": List[str], "allow_none": True},
    "global_megacomplex_scale": {"type": List[Parameter], "default": None, "allow_none": True},
    "scale": {"type": Parameter, "default": None, "allow_none": True},
}


class Model:
    """A base class for global analysis models."""

    def __init__(
        self,
        *,
        megacomplex_types: dict[str, type[Megacomplex]],
        default_megacomplex_type: str | None = None,
        dataset_group_models: dict[str, DatasetGroupModel] = None,
    ):
        self._megacomplex_types = megacomplex_types
        self._default_megacomplex_type = default_megacomplex_type or next(iter(megacomplex_types))

        self._dataset_group_models = dataset_group_models or {"default": DatasetGroupModel()}
        if "default" not in self._dataset_group_models:
            self._dataset_group_models["default"] = DatasetGroupModel()

        self._model_items = {}
        self._dataset_properties = {}
        self._add_default_items_and_properties()
        self._add_megacomplexe_types()
        self._add_dataset_type()

    @classmethod
    def from_dict(
        cls,
        model_dict: dict[str, Any],
        *,
        megacomplex_types: dict[str, type[Megacomplex]] | None = None,
        default_megacomplex_type: str | None = None,
    ) -> Model:
        """Creates a model from a dictionary.

        Parameters
        ----------
        model_dict: dict[str, Any]
            Dictionary containing the model.
        megacomplex_types: dict[str, type[Megacomplex]] | None
            Overwrite 'megacomplex_types' in ``model_dict`` for testing.
        default_megacomplex_type: str | None
            Overwrite 'default_megacomplex' in ``model_dict`` for testing.
        """
        model_dict = copy.deepcopy(model_dict)
        if default_megacomplex_type is None:
            default_megacomplex_type = model_dict.get("default_megacomplex")

        if megacomplex_types is None:
            megacomplex_types = {
                m["type"]: get_megacomplex(m["type"])
                for m in model_dict["megacomplex"].values()
                if "type" in m
            }
        if (
            default_megacomplex_type is not None
            and default_megacomplex_type not in megacomplex_types
        ):
            megacomplex_types[default_megacomplex_type] = get_megacomplex(default_megacomplex_type)
        if "default_megacomplex" in model_dict:
            model_dict.pop("default_megacomplex", None)

        dataset_group_models = model_dict.pop("dataset_groups", None)
        if dataset_group_models is not None:
            dataset_group_models = {
                label: DatasetGroupModel(**group) for label, group in dataset_group_models.items()
            }

        model = cls(
            megacomplex_types=megacomplex_types,
            default_megacomplex_type=default_megacomplex_type,
            dataset_group_models=dataset_group_models,
        )

        # iterate over items
        for item_name, items in list(model_dict.items()):

            if item_name not in model.model_items:
                warn(f"Unknown model item type '{item_name}'.")
                continue

            is_list = isinstance(getattr(model, item_name), list)

            if is_list:
                model._add_list_items(item_name, items)
            else:
                model._add_dict_items(item_name, items)

        return model

    def _add_dict_items(self, item_name: str, items: dict):

        for label, item in items.items():
            item_cls = self.model_items[item_name]
            is_typed = hasattr(item_cls, "_glotaran_model_item_typed")
            if is_typed:
                if "type" not in item and item_cls.get_default_type() is None:
                    raise ValueError(f"Missing type for attribute '{item_name}'")
                item_type = item.get("type", item_cls.get_default_type())

                types = item_cls._glotaran_model_item_types
                if item_type not in types:
                    raise ValueError(f"Unknown type '{item_type}' for attribute '{item_name}'")
                item_cls = types[item_type]
            item["label"] = label
            item = item_cls.from_dict(item)
            getattr(self, item_name)[label] = item

    def _add_list_items(self, item_name: str, items: list):

        for item in items:
            item_cls = self.model_items[item_name]
            is_typed = hasattr(item_cls, "_glotaran_model_item_typed")
            if is_typed:
                if "type" not in item:
                    raise ValueError(f"Missing type for attribute '{item_name}'")
                item_type = item["type"]

                if item_type not in item_cls._glotaran_model_item_types:
                    raise ValueError(f"Unknown type '{item_type}' for attribute '{item_name}'")
                item_cls = item_cls._glotaran_model_item_types[item_type]
            item = item_cls.from_dict(item)
            getattr(self, item_name).append(item)

    def _add_megacomplexe_types(self):

        for megacomplex_name, megacomplex_type in self._megacomplex_types.items():
            if not issubclass(megacomplex_type, Megacomplex):
                raise TypeError(
                    f"Megacomplex type {megacomplex_name}({megacomplex_type}) "
                    "is not a subclass of Megacomplex"
                )
            self._add_megacomplex_type(megacomplex_type)

        model_megacomplex_type = create_model_megacomplex_type(
            self._megacomplex_types, self.default_megacomplex
        )
        self._add_model_item("megacomplex", model_megacomplex_type)

    def _add_megacomplex_type(self, megacomplex_type: type[Megacomplex]):

        for item_name, item in megacomplex_type.glotaran_model_items().items():
            self._add_model_item(item_name, item)

        for item_name, item in megacomplex_type.glotaran_dataset_model_items().items():
            self._add_model_item(item_name, item)

        for property_name, prop in megacomplex_type.glotaran_dataset_properties().items():
            self._add_dataset_property(property_name, prop)

    def _add_model_item(self, item_name: str, item: type):
        if item_name in self._model_items:
            if self.model_items[item_name] != item:
                raise ModelError(
                    f"Cannot add item of type {item_name}. Model item '{item_name}' "
                    "was already defined as a different type."
                )
            return
        self._model_items[item_name] = item

        if getattr(item, "_glotaran_has_label"):
            setattr(self, f"{item_name}", {})
        else:
            setattr(self, f"{item_name}", [])

    def _add_dataset_property(self, property_name: str, dataset_property: dict[str, any]):
        if property_name in self._dataset_properties:
            known_type = (
                self._dataset_properties[property_name]
                if not isinstance(self._dataset_properties, dict)
                else self._dataset_properties[property_name]["type"]
            )
            new_type = (
                dataset_property
                if not isinstance(dataset_property, dict)
                else dataset_property["type"]
            )
            if known_type != new_type:
                raise ModelError(
                    f"Cannot add dataset property of type {property_name} as it was "
                    "already defined as a different type."
                )
            return
        self._dataset_properties[property_name] = dataset_property

    def _add_default_items_and_properties(self):
        for item_name, item in default_model_items.items():
            self._add_model_item(item_name, item)

        for property_name, prop in default_dataset_properties.items():
            self._add_dataset_property(property_name, prop)

    def _add_dataset_type(self):
        dataset_model_type = create_dataset_model_type(self._dataset_properties)
        self._add_model_item("dataset", dataset_model_type)

    @property
    def model_dimension(self):
        """Deprecated use ``Scheme.model_dimensions['<dataset_name>']`` instead"""
        raise_deprecation_error(
            deprecated_qual_name_usage="Model.model_dimension",
            new_qual_name_usage=("Scheme.model_dimensions['<dataset_name>']"),
            to_be_removed_in_version="0.7.0",
        )

    @property
    def global_dimension(self):
        """Deprecated use ``Scheme.global_dimensions['<dataset_name>']`` instead"""
        raise_deprecation_error(
            deprecated_qual_name_usage="Model.global_dimension",
            new_qual_name_usage=("Scheme.global_dimensions['<dataset_name>']"),
            to_be_removed_in_version="0.7.0",
        )

    @property
    def default_megacomplex(self) -> str:
        """The default megacomplex used by this model."""
        return self._default_megacomplex_type

    @property
    def megacomplex_types(self) -> dict[str, type[Megacomplex]]:
        """The megacomplex types used by this model."""
        return self._megacomplex_types

    @property
    def dataset_group_models(self) -> dict[str, DatasetGroupModel]:
        return self._dataset_group_models

    @property
    def model_items(self) -> dict[str, type[object]]:
        """The model_items types used by this model."""
        return self._model_items

    @property
    def global_megacomplex(self) -> dict[str, Megacomplex]:
        """Alias for `glotaran.model.megacomplex`. Needed internally."""
        return self.megacomplex

    def get_dataset_groups(self) -> dict[str, DatasetGroup]:
        groups = {}
        for dataset_model in self.dataset.values():
            group = dataset_model.group
            if group not in groups:
                try:
                    groups[group] = DatasetGroup(model=self.dataset_group_models[group])
                except KeyError:
                    raise ValueError(f"Unknown dataset group '{group}'")
            groups[group].dataset_models[dataset_model.label] = dataset_model
        return groups

    def as_dict(self) -> dict:
        model_dict = {
            "default_megacomplex": self.default_megacomplex,
            "dataset_groups": {
                label: asdict(group) for label, group in self.dataset_group_models.items()
            },
        }
        for item_name in self._model_items:
            items = getattr(self, item_name)
            if len(items) == 0:
                continue
            if isinstance(items, list):
                model_dict[item_name] = [item.as_dict() for item in items]
            else:
                model_dict[item_name] = {label: item.as_dict() for label, item in items.items()}

        return model_dict

    def get_parameters(self) -> list[str]:
        parameters = []
        for item_name in self.model_items:
            items = getattr(self, item_name)
            item_iterator = items if isinstance(items, list) else items.values()
            for item in item_iterator:
                parameters += item.get_parameters()
        return parameters

    def need_index_dependent(self) -> bool:
        """Returns true if e.g. clp_relations with intervals are present."""
        return any(i.interval is not None for i in self.clp_constraints + self.clp_relations)

    def is_groupable(self, parameters: ParameterGroup, data: dict[str, xr.DataArray]) -> bool:
        dataset_models = {label: self.dataset[label] for label in data}
        if any(d.has_global_model() for d in dataset_models.values()):
            return False
        global_dimensions = {
            d.fill(self, parameters).set_data(data[k]).get_global_dimension()
            for k, d in dataset_models.items()
        }
        model_dimensions = {
            d.fill(self, parameters).set_data(data[k]).get_model_dimension()
            for k, d in dataset_models.items()
        }
        return len(global_dimensions) == 1 and len(model_dimensions) == 1

    def problem_list(self, parameters: ParameterGroup = None) -> list[str]:
        """
        Returns a list with all problems in the model and missing parameters if specified.

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        problems = []

        for name in self.model_items:
            items = getattr(self, name)
            if isinstance(items, list):
                for item in items:
                    problems += item.validate(self, parameters=parameters)
            else:
                for _, item in items.items():
                    problems += item.validate(self, parameters=parameters)

        return problems

    def validate(self, parameters: ParameterGroup = None, raise_exception: bool = False) -> str:
        """
        Returns a string listing all problems in the model and missing parameters if specified.

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        result = ""

        problems = self.problem_list(parameters)
        if problems:
            result = f"Your model has {len(problems)} problems:\n"
            for p in problems:
                result += f"\n * {p}"
            if raise_exception:
                raise ModelError(result)
        else:
            result = "Your model is valid."
        return result

    def valid(self, parameters: ParameterGroup = None) -> bool:
        """Returns `True` if the number problems in the model is 0, else `False`

        Parameters
        ----------

        parameter :
            The parameter to validate.
        """
        return len(self.problem_list(parameters)) == 0

    def markdown(
        self,
        parameters: ParameterGroup = None,
        initial_parameters: ParameterGroup = None,
        base_heading_level: int = 1,
    ) -> MarkdownStr:
        """Formats the model as Markdown string.

        Parameters will be included if specified.

        Parameters
        ----------
        parameter: ParameterGroup
            Parameter to include.
        initial_parameters: ParameterGroup
            Initial values for the parameters.
        base_heading_level: int
            Base heading level of the markdown sections.

            E.g.:

            - If it is 1 the string will start with '# Model'.
            - If it is 3 the string will start with '### Model'.
        """
        base_heading = "#" * base_heading_level
        string = f"{base_heading} Model\n\n"
        string += "_Megacomplex Types_: "
        string += ", ".join(self._megacomplex_types)
        string += "\n\n"

        for name in self.model_items:
            items = getattr(self, name)
            if not items:
                continue

            string += f"{base_heading}# {name.replace('_', ' ').title()}\n\n"

            if isinstance(items, dict):
                items = items.values()
            for item in items:
                item_str = item.mprint(
                    parameters=parameters, initial_parameters=initial_parameters
                ).split("\n")
                string += f"* {item_str[0]}\n"
                for s in item_str[1:]:
                    string += f"  {s}\n"
            string += "\n"
        return MarkdownStr(string)

    def _repr_markdown_(self) -> str:
        """Special method used by ``ipython`` to render markdown."""
        return str(self.markdown(base_heading_level=3))

    def __str__(self) -> str:
        return str(self.markdown())
